/*----- PROTECTED REGION ID(CCDSpectrometer.h) ENABLED START -----*/
/* clang-format on */
//=============================================================================
//
// file :        CCDSpectrometer.h
//
// description : Include file for the CCDSpectrometer class
//
// project :     CCDSpectrometer
//
// This file is part of Tango device class.
//
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
//
//
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#ifndef CCDSpectrometer_H
#define CCDSpectrometer_H

#include <tango.h>
#include "readthread.h"
#include "Fit.h"
#include "tbuffer.hpp"

#define MAX_GAUSS_ITERATION     25
#define MAX_CONFITEOR_ITERATION 20
#define NUM_GAUSS_PARAMETER     4
#define AUTO_EXPOSURE_MEAN  4
#define SIGMA_TO_FWHM_FACTOR    0.4246609

#define NUM_BUF		5
#define BUF_SIZE			10000
#define BUF_ID_FWHM_CENTER		0
#define BUF_ID_FWHM_WIDTH		1
#define BUF_ID_GAUSS_CENTER		2
#define BUF_ID_GAUSS_WIDTH		3
#define BUF_ID_AREA				4

#define DATA_ATTR				"HorProfile"
#define EXPOSURE_ATTR			"Exposure"
#define TRIGGER_ATTR			"Trigger"
#define COUNTER_ATTR			"AcquisitionCounter"
#define BUNCHNUM_ATTR			"BunchNumber"
#define ROI_ATTR				"HwRoiParam"
#define START_CMD				"Start"
#define STOP_CMD				"Stop"
#define GET_DATA_CMD			"GetHorPosProfile"

/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CCDSpectrometer.h

#ifdef TANGO_LOG
	// cppTango after c934adea (Merge branch 'remove-cout-definition' into 'main', 2022-05-23)
	// nothing to do
#else
	// cppTango 9.3-backports and older
	#define TANGO_LOG       cout
	#define TANGO_LOG_INFO  cout2
	#define TANGO_LOG_DEBUG cout3
#endif // TANGO_LOG

/**
 *  CCDSpectrometer class description:
 *    Laser Spectrometer based on Basler CCD Tango Device
 */


namespace CCDSpectrometer_ns
{
/*----- PROTECTED REGION ID(CCDSpectrometer::Additional Class Declarations) ENABLED START -----*/
/* clang-format on */
//	Additional Class Declarations
uint16_t const    MAX_NR_PIXELS           = 4096;
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CCDSpectrometer::Additional Class Declarations

class CCDSpectrometer : public TANGO_BASE_CLASS
{

/*----- PROTECTED REGION ID(CCDSpectrometer::Data Members) ENABLED START -----*/
/* clang-format on */
//	Add your own data members
	readthread *loop;
public:
	Tango::DeviceProxy *dp;
	omni_mutex *dpmutex;
	tbuffer::buffer_t<double> tbuf[NUM_BUF];
	tbuffer::buffer_t<tbuffer::_data_array<double> > tbuf_array;//GetData
	tbuffer::buffer_t<tbuffer::_data_array<double> > tbuf_array_w;//GetWave
	tbuffer::buffer_t<tbuffer::_data_array<double> > tbuf_array_l;//GetLambda
	
	bool conn_status;
	bool abortflag;
	bool live_mode;
	bool init_live_mode;
	bool measure_mode;
	bool init_measure_mode;
	bool acquire_noise_mode;
	bool init_acquire_noise_mode;
	bool stop_mode;

	uint32_t noise_frames;
	int noise_frames_counter;
	int noise_level;
	timespec last_read;

	/*LibFit_ns::*/Fit *m_gaussfit;//
	/*LibFit_ns::*/Fit *m_asyngaussfit;//
	/*LibFit_ns::*/Fit *m_confiteorfit;//
	bool GaussFit_valid;

	//DWORD *pMeasData;
	double *pLiveData;
	double *pLiveData_avg;
	double *pNoiseMask_avg;
	double *pFrameBuffer;		//bi-dimensional array
	double *pReferenceData;
	unsigned int FrameBufferIndex;
	Tango::DevDouble *x_values;
	unsigned int RegionOfInterestStartPix;
	unsigned int RegionOfInterestEndPix;
	unsigned int AnalysisRegionStartPix;
	unsigned int AnalysisRegionEndPix;
	void init_data(bool reconnection);
	unsigned int WaveLenght2Pix(double val);
	bool fwhm_calculate(double */*restrict*/ data_array, size_t dim_frame, size_t x_start_analysis, size_t x_end_analisys, size_t num_frames, size_t avg_val, size_t &x_peak, double &y_peak, size_t &x_half_min, size_t &x_half_max);
	int y_max_mean;
	int exposures[AUTO_EXPOSURE_MEAN];
	int exposure_idx;
	void string_explode(string str, string separator, vector<string>* results);
	void load_intensity_cal_file(string filename);
	void save_intensity_cal_file(string filename);
	uint32_t myntohl(uint32_t val_l);

	float fit[5/*NR_WAVELEN_POL_COEF*/];

	double m_pLambda[MAX_NR_PIXELS];
	unsigned int l_Time;
	double l_pSpectrum[MAX_NR_PIXELS];//TODO: REMOVE
	char   attr_strchar[1024];
	Tango::DevString attr_tmp_ptr;
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CCDSpectrometer::Data Members

//	Device property data members
public:
	//	DeviceName:	
	std::string	deviceName;
	//	DefaultGaussRef:	Default Gauss Reference
	std::vector<Tango::DevDouble>	defaultGaussRef;
	//	DataReferenceFile:	Default Data Reference File
	std::string	dataReferenceFile;
	//	DefaultStartRegion:	Beginning of the region of interest
	Tango::DevDouble	defaultStartRegion;
	//	DefaultEndRegion:	End of the region of interest
	Tango::DevDouble	defaultEndRegion;
	//	PreferFWHM:	Choose FWHM instead og GaussianFit if both enabled
	Tango::DevBoolean	preferFWHM;
	//	AutoExposureRange:	AutoExposure Range
	std::vector<Tango::DevDouble>	autoExposureRange;
	//	AutoExposureStep:	AutoExposure Step
	Tango::DevDouble	autoExposureStep;
	//	AutoExposureAmpRange:	AutoExposure Amp Range
	std::vector<Tango::DevDouble>	autoExposureAmpRange;
	//	IntensityCalibrationFile:	
	std::string	intensityCalibrationFile;
	//	NoiseProfileFile:	
	std::string	noiseProfileFile;
	//	WavelengthAttr:	device name
	//  attribute name
	std::vector<std::string>	wavelengthAttr;
	//	AreaInAnalysisRegion:	use AnalysisRegion for area instead of AreaRegion
	Tango::DevBoolean	areaInAnalysisRegion;
	//	WidthCorrection:	widht_in^2 + widthcorrection^2 = widht_out^2
	Tango::DevDouble	widthCorrection;
	//	NFramesBuffered:	number of frames buffered when in live mode
	Tango::DevLong	nFramesBuffered;
	//	AnalysisRegion:	AnalysisRegion
	std::vector<Tango::DevDouble>	analysisRegion;
	//	RegionOfInterest:	RegionOfInterest
	std::vector<Tango::DevDouble>	regionOfInterest;
	//	ExposureTime:	ExposureTime
	Tango::DevDouble	exposureTime;
	//	SyncMode:	ExposureTime
	Tango::DevLong	syncMode;
	//	AutoStart:	AutoStart
	Tango::DevBoolean	autoStart;
	//	NumPixel:	Number of pixel of the CCD
	Tango::DevLong	numPixel;
	//	CameraLabel:	Camera label or IP
	std::string	cameraLabel;

//	Attribute data members
public:
	Tango::DevLong	*attr_SyncMode_read;
	Tango::DevDouble	*attr_ExposureTime_read;
	Tango::DevBoolean	*attr_NoiseSubtraction_read;
	Tango::DevBoolean	*attr_GaussFitEnable_read;
	Tango::DevBoolean	*attr_FWHMEnable_read;
	Tango::DevDouble	*attr_GaussFitCenter_read;
	Tango::DevDouble	*attr_GaussFitWidth_read;
	Tango::DevDouble	*attr_FWHMCenter_read;
	Tango::DevDouble	*attr_FWHMWidth_read;
	Tango::DevDouble	*attr_Center_read;
	Tango::DevDouble	*attr_Width_read;
	Tango::DevLong	*attr_FitType_read;
	Tango::DevLong	*attr_FWHMAvg_read;
	Tango::DevLong	*attr_NoiseLevel_read;
	Tango::DevString	*attr_DataReferenceFile_read;
	Tango::DevBoolean	*attr_AutoExposure_read;
	Tango::DevBoolean	*attr_RtMode_read;
	Tango::DevBoolean	*attr_IntensityCalibrationEnable_read;
	Tango::DevDouble	*attr_Area_read;
	Tango::DevDouble	*attr_AreaRegion_read;
	Tango::DevBoolean	*attr_AreaEnable_read;
	Tango::DevBoolean	*attr_WidthCorrectionEnable_read;
	Tango::DevBoolean	*attr_IncrSensitivity_read;
	Tango::DevDouble	*attr_Data_read;
	Tango::DevLong	*attr_NumPixels_read;
	Tango::DevDouble	*attr_RegionOfInterest_read;
	Tango::DevDouble	*attr_AnalysisRegion_read;
	Tango::DevDouble	*attr_GaussFit_read;
	Tango::DevDouble	*attr_FWHM_read;
	Tango::DevDouble	*attr_GaussReference_read;
	Tango::DevDouble	*attr_DataReference_read;
	Tango::DevDouble	*attr_IntensityCalibration_read;
	Tango::DevDouble	*attr_NoiseProfile_read;
	Tango::DevDouble	*attr_Wave_read;
	Tango::DevDouble	*attr_Lambda_read;

//	Constructors and destructors
public:
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device Name
	 */
	CCDSpectrometer(Tango::DeviceClass *cl,std::string &s);
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device Name
	 */
	CCDSpectrometer(Tango::DeviceClass *cl,const char *s);
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device name
	 *	@param d	Device description.
	 */
	CCDSpectrometer(Tango::DeviceClass *cl,const char *s,const char *d);
	/**
	 * The device object destructor.
	 */
	~CCDSpectrometer();


//	Miscellaneous methods
public:
	/*
	 *	will be called at device destruction or at init command.
	 */
	void delete_device();
	/*
	 *	Initialize the device
	 */
	virtual void init_device();
	/*
	 *	Read the device properties from database
	 */
	void get_device_property();
	/*
	 *	Always executed method before execution command method.
	 */
	virtual void always_executed_hook();


//	Attribute methods
public:
	//--------------------------------------------------------
	/*
	 *	Method     : CCDSpectrometer::read_attr_hardware()
	 *	Description: Hardware acquisition for attributes.
	 */
	//--------------------------------------------------------
	virtual void read_attr_hardware(std::vector<long> &attr_list);
	//--------------------------------------------------------
	/*
	 *	Method     : CCDSpectrometer::write_attr_hardware()
	 *	Description: Hardware writing for attributes.
	 */
	//--------------------------------------------------------
	virtual void write_attr_hardware(std::vector<long> &attr_list);

/**
 *	Attribute SyncMode related methods
 *	Description: 0=SW
 *               1=HW Sync input
 *               2=HW external trigger
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
	virtual void read_SyncMode(Tango::Attribute &attr);
	virtual void write_SyncMode(Tango::WAttribute &attr);
	virtual bool is_SyncMode_allowed(Tango::AttReqType type);
/**
 *	Attribute ExposureTime related methods
 *	Description: Exposure time in milliseconds
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_ExposureTime(Tango::Attribute &attr);
	virtual void write_ExposureTime(Tango::WAttribute &attr);
	virtual bool is_ExposureTime_allowed(Tango::AttReqType type);
/**
 *	Attribute NoiseSubtraction related methods
 *	Description: Enable/Disable subtraction of background noise profile 
 *               acquired with the AcquireNoise command
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_NoiseSubtraction(Tango::Attribute &attr);
	virtual void write_NoiseSubtraction(Tango::WAttribute &attr);
	virtual bool is_NoiseSubtraction_allowed(Tango::AttReqType type);
/**
 *	Attribute GaussFitEnable related methods
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_GaussFitEnable(Tango::Attribute &attr);
	virtual void write_GaussFitEnable(Tango::WAttribute &attr);
	virtual bool is_GaussFitEnable_allowed(Tango::AttReqType type);
/**
 *	Attribute FWHMEnable related methods
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_FWHMEnable(Tango::Attribute &attr);
	virtual void write_FWHMEnable(Tango::WAttribute &attr);
	virtual bool is_FWHMEnable_allowed(Tango::AttReqType type);
/**
 *	Attribute GaussFitCenter related methods
 *	Description: wavelength of peak of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_GaussFitCenter(Tango::Attribute &attr);
	virtual bool is_GaussFitCenter_allowed(Tango::AttReqType type);
/**
 *	Attribute GaussFitWidth related methods
 *	Description: width of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_GaussFitWidth(Tango::Attribute &attr);
	virtual bool is_GaussFitWidth_allowed(Tango::AttReqType type);
/**
 *	Attribute FWHMCenter related methods
 *	Description: wavelength of peak of FWHM
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_FWHMCenter(Tango::Attribute &attr);
	virtual bool is_FWHMCenter_allowed(Tango::AttReqType type);
/**
 *	Attribute FWHMWidth related methods
 *	Description: width of FWHM
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_FWHMWidth(Tango::Attribute &attr);
	virtual bool is_FWHMWidth_allowed(Tango::AttReqType type);
/**
 *	Attribute Center related methods
 *	Description: wavelength of peak of FWHM or of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_Center(Tango::Attribute &attr);
	virtual bool is_Center_allowed(Tango::AttReqType type);
/**
 *	Attribute Width related methods
 *	Description: width of FWHM or of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_Width(Tango::Attribute &attr);
	virtual bool is_Width_allowed(Tango::AttReqType type);
/**
 *	Attribute FitType related methods
 *	Description: 1=Gaussian 2=Gaussian Asymm 3=Confiteor
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
	virtual void read_FitType(Tango::Attribute &attr);
	virtual void write_FitType(Tango::WAttribute &attr);
	virtual bool is_FitType_allowed(Tango::AttReqType type);
/**
 *	Attribute FWHMAvg related methods
 *
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
	virtual void read_FWHMAvg(Tango::Attribute &attr);
	virtual void write_FWHMAvg(Tango::WAttribute &attr);
	virtual bool is_FWHMAvg_allowed(Tango::AttReqType type);
/**
 *	Attribute NoiseLevel related methods
 *
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
	virtual void read_NoiseLevel(Tango::Attribute &attr);
	virtual void write_NoiseLevel(Tango::WAttribute &attr);
	virtual bool is_NoiseLevel_allowed(Tango::AttReqType type);
/**
 *	Attribute DataReferenceFile related methods
 *
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
	virtual void read_DataReferenceFile(Tango::Attribute &attr);
	virtual void write_DataReferenceFile(Tango::WAttribute &attr);
	virtual bool is_DataReferenceFile_allowed(Tango::AttReqType type);
/**
 *	Attribute AutoExposure related methods
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_AutoExposure(Tango::Attribute &attr);
	virtual void write_AutoExposure(Tango::WAttribute &attr);
	virtual bool is_AutoExposure_allowed(Tango::AttReqType type);
/**
 *	Attribute RtMode related methods
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_RtMode(Tango::Attribute &attr);
	virtual void write_RtMode(Tango::WAttribute &attr);
	virtual bool is_RtMode_allowed(Tango::AttReqType type);
/**
 *	Attribute IntensityCalibrationEnable related methods
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_IntensityCalibrationEnable(Tango::Attribute &attr);
	virtual void write_IntensityCalibrationEnable(Tango::WAttribute &attr);
	virtual bool is_IntensityCalibrationEnable_allowed(Tango::AttReqType type);
/**
 *	Attribute Area related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_Area(Tango::Attribute &attr);
	virtual bool is_Area_allowed(Tango::AttReqType type);
/**
 *	Attribute AreaRegion related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_AreaRegion(Tango::Attribute &attr);
	virtual void write_AreaRegion(Tango::WAttribute &attr);
	virtual bool is_AreaRegion_allowed(Tango::AttReqType type);
/**
 *	Attribute AreaEnable related methods
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_AreaEnable(Tango::Attribute &attr);
	virtual void write_AreaEnable(Tango::WAttribute &attr);
	virtual bool is_AreaEnable_allowed(Tango::AttReqType type);
/**
 *	Attribute WidthCorrectionEnable related methods
 *	Description: enable widht correction
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_WidthCorrectionEnable(Tango::Attribute &attr);
	virtual void write_WidthCorrectionEnable(Tango::WAttribute &attr);
	virtual bool is_WidthCorrectionEnable_allowed(Tango::AttReqType type);
/**
 *	Attribute IncrSensitivity related methods
 *	Description: false: use standard sensitivity
 *               true: use increased sensitivity
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_IncrSensitivity(Tango::Attribute &attr);
	virtual void write_IncrSensitivity(Tango::WAttribute &attr);
	virtual bool is_IncrSensitivity_allowed(Tango::AttReqType type);
/**
 *	Attribute Data related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4100
 */
	virtual void read_Data(Tango::Attribute &attr);
	virtual bool is_Data_allowed(Tango::AttReqType type);
/**
 *	Attribute NumPixels related methods
 *	Description: long[0]: number of pixels horizontally (number of CCD-array columns)
 *               long[1]: number of pixels vertically (number of CCD-array lines)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Spectrum max = 2
 */
	virtual void read_NumPixels(Tango::Attribute &attr);
	virtual bool is_NumPixels_allowed(Tango::AttReqType type);
/**
 *	Attribute RegionOfInterest related methods
 *	Description: Region of interest returned by the Data attribute
 *               double[0]=start of the region
 *               double[1]=end of the region
 *               
 *               Values are in nm according with the conversion factor in x_values
 *               Data attribute property
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 2
 */
	virtual void read_RegionOfInterest(Tango::Attribute &attr);
	virtual void write_RegionOfInterest(Tango::WAttribute &attr);
	virtual bool is_RegionOfInterest_allowed(Tango::AttReqType type);
/**
 *	Attribute AnalysisRegion related methods
 *	Description: Region used for gaussian and fwhm analysis
 *               double[0]=start of the region
 *               double[1]=end of the region
 *               
 *               Values are in nm according with the conversion factor in x_values
 *               Data attribute property
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 2
 */
	virtual void read_AnalysisRegion(Tango::Attribute &attr);
	virtual void write_AnalysisRegion(Tango::WAttribute &attr);
	virtual bool is_AnalysisRegion_allowed(Tango::AttReqType type);
/**
 *	Attribute GaussFit related methods
 *	Description: double[0]= wavelength of peak of Gaussian Fit
 *               double[1]= width of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
	virtual void read_GaussFit(Tango::Attribute &attr);
	virtual bool is_GaussFit_allowed(Tango::AttReqType type);
/**
 *	Attribute FWHM related methods
 *	Description: double[0]= wavelength of peak of FWHM
 *               double[1]= width of FWHM
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 3
 */
	virtual void read_FWHM(Tango::Attribute &attr);
	virtual bool is_FWHM_allowed(Tango::AttReqType type);
/**
 *	Attribute GaussReference related methods
 *	Description: double[0]= wavelength of peak of Gaussian Fit
 *               double[1]= width of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
	virtual void read_GaussReference(Tango::Attribute &attr);
	virtual void write_GaussReference(Tango::WAttribute &attr);
	virtual bool is_GaussReference_allowed(Tango::AttReqType type);
/**
 *	Attribute DataReference related methods
 *	Description: A reference curve
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
	virtual void read_DataReference(Tango::Attribute &attr);
	virtual void write_DataReference(Tango::WAttribute &attr);
	virtual bool is_DataReference_allowed(Tango::AttReqType type);
/**
 *	Attribute IntensityCalibration related methods
 *	Description: A reference curve
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
	virtual void read_IntensityCalibration(Tango::Attribute &attr);
	virtual void write_IntensityCalibration(Tango::WAttribute &attr);
	virtual bool is_IntensityCalibration_allowed(Tango::AttReqType type);
/**
 *	Attribute NoiseProfile related methods
 *	Description: A reference curve
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
	virtual void read_NoiseProfile(Tango::Attribute &attr);
	virtual bool is_NoiseProfile_allowed(Tango::AttReqType type);
/**
 *	Attribute Wave related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
	virtual void read_Wave(Tango::Attribute &attr);
	virtual bool is_Wave_allowed(Tango::AttReqType type);
/**
 *	Attribute Lambda related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
	virtual void read_Lambda(Tango::Attribute &attr);
	virtual bool is_Lambda_allowed(Tango::AttReqType type);


	//--------------------------------------------------------
	/**
	 *	Method     : CCDSpectrometer::add_dynamic_attributes()
	 *	Description: Add dynamic attributes if any.
	 */
	//--------------------------------------------------------
	void add_dynamic_attributes();




//	Command related methods
public:
	/**
	 *	Command Start related method
	 *
	 *
	 */
	virtual void start();
	virtual bool is_Start_allowed(const CORBA::Any &any);
	/**
	 *	Command Stop related method
	 *
	 *
	 */
	virtual void stop();
	virtual bool is_Stop_allowed(const CORBA::Any &any);
	/**
	 *	Command Reset related method
	 *
	 *
	 */
	virtual void reset();
	virtual bool is_Reset_allowed(const CORBA::Any &any);
	/**
	 *	Command StartSingle related method
	 *	Description: Start single acquisition
	 *
	 */
	virtual void start_single();
	virtual bool is_StartSingle_allowed(const CORBA::Any &any);
	/**
	 *	Command StartContinuous related method
	 *	Description: Start acquisition in Continuous mode
	 *
	 */
	virtual void start_continuous();
	virtual bool is_StartContinuous_allowed(const CORBA::Any &any);
	/**
	 *	Command GetLastData related method
	 *	Description: returns the last argin frames saved
	 *
	 *	@param argin number of frames to return
	 *	@returns
	 */
	virtual Tango::DevVarDoubleArray *get_last_data(Tango::DevLong argin);
	virtual bool is_GetLastData_allowed(const CORBA::Any &any);
	/**
	 *	Command AcquireNoise related method
	 *	Description: Acquire argin frames of background noise
	 *
	 *	@param argin Number of frames to acquire as noise
	 */
	virtual void acquire_noise(Tango::DevLong argin);
	virtual bool is_AcquireNoise_allowed(const CORBA::Any &any);
	/**
	 *	Command SaveReferenceData related method
	 *	Description: Save current curve as reference curve
	 *
	 */
	virtual void save_reference_data();
	virtual bool is_SaveReferenceData_allowed(const CORBA::Any &any);
	/**
	 *	Command GetData related method
	 *
	 *
	 *	@param argin
	 *	@returns
	 */
	virtual Tango::DevVarDoubleArray *get_data(const Tango::DevVarLongArray *argin);
	virtual bool is_GetData_allowed(const CORBA::Any &any);
	/**
	 *	Command GetGaussFitCenter related method
	 *
	 *
	 *	@param argin
	 *	@returns
	 */
	virtual Tango::DevVarDoubleArray *get_gauss_fit_center(const Tango::DevVarLongArray *argin);
	virtual bool is_GetGaussFitCenter_allowed(const CORBA::Any &any);
	/**
	 *	Command GetGaussFitWidth related method
	 *
	 *
	 *	@param argin
	 *	@returns
	 */
	virtual Tango::DevVarDoubleArray *get_gauss_fit_width(const Tango::DevVarLongArray *argin);
	virtual bool is_GetGaussFitWidth_allowed(const CORBA::Any &any);
	/**
	 *	Command GetFWHMCenter related method
	 *
	 *
	 *	@param argin
	 *	@returns
	 */
	virtual Tango::DevVarDoubleArray *get_fwhmcenter(const Tango::DevVarLongArray *argin);
	virtual bool is_GetFWHMCenter_allowed(const CORBA::Any &any);
	/**
	 *	Command GetFWHMWidth related method
	 *
	 *
	 *	@param argin
	 *	@returns
	 */
	virtual Tango::DevVarDoubleArray *get_fwhmwidth(const Tango::DevVarLongArray *argin);
	virtual bool is_GetFWHMWidth_allowed(const CORBA::Any &any);
	/**
	 *	Command ExportWavelength related method
	 *
	 *
	 */
	virtual void export_wavelength();
	virtual bool is_ExportWavelength_allowed(const CORBA::Any &any);
	/**
	 *	Command GetArea related method
	 *
	 *
	 *	@param argin
	 *	@returns
	 */
	virtual Tango::DevVarDoubleArray *get_area(const Tango::DevVarLongArray *argin);
	virtual bool is_GetArea_allowed(const CORBA::Any &any);
	/**
	 *	Command GetWave related method
	 *	Description: Same as GetData
	 *
	 *	@param argin
	 *	@returns
	 */
	virtual Tango::DevVarDoubleArray *get_wave(const Tango::DevVarLongArray *argin);
	virtual bool is_GetWave_allowed(const CORBA::Any &any);
	/**
	 *	Command GetLambda related method
	 *
	 *
	 *	@param argin
	 *	@returns
	 */
	virtual Tango::DevVarDoubleArray *get_lambda(const Tango::DevVarLongArray *argin);
	virtual bool is_GetLambda_allowed(const CORBA::Any &any);


	//--------------------------------------------------------
	/**
	 *	Method     : CCDSpectrometer::add_dynamic_commands()
	 *	Description: Add dynamic commands if any.
	 */
	//--------------------------------------------------------
	void add_dynamic_commands();

/*----- PROTECTED REGION ID(CCDSpectrometer::Additional Method prototypes) ENABLED START -----*/
/* clang-format on */
//	Additional Method prototypes

/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CCDSpectrometer::Additional Method prototypes
};

/*----- PROTECTED REGION ID(CCDSpectrometer::Additional Classes Definitions) ENABLED START -----*/
/* clang-format on */
//	Additional Classes Definitions
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CCDSpectrometer::Additional Classes Definitions

}	//	End of namespace

#endif   //	CCDSpectrometer_H
