/*----- PROTECTED REGION ID(CCDSpectrometerClass.cpp) ENABLED START -----*/
/* clang-format on */
//=============================================================================
//
// file :        CCDSpectrometerClass.cpp
//
// description : C++ source for the CCDSpectrometerClass.
//               A singleton class derived from DeviceClass.
//               It implements the command and attribute list
//               and all properties and methods required
//               by the CCDSpectrometer once per process.
//
// project :     CCDSpectrometer
//
// This file is part of Tango device class.
//
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
//
//
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include "CCDSpectrometerClass.h"
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CCDSpectrometerClass.cpp

//-------------------------------------------------------------------
/**
 *	Create CCDSpectrometerClass singleton and
 *	return it in a C function for Python usage
 */
//-------------------------------------------------------------------
extern "C" {
#ifdef _TG_WINDOWS_

__declspec(dllexport)

#endif

	Tango::DeviceClass *_create_CCDSpectrometer_class(const char *name) {
		return CCDSpectrometer_ns::CCDSpectrometerClass::init(name);
	}
}

namespace CCDSpectrometer_ns
{
//===================================================================
//	Initialize pointer for singleton pattern
//===================================================================
CCDSpectrometerClass *CCDSpectrometerClass::_instance = NULL;

//===================================================================
//	Class constants
//===================================================================
const long DataAttrib::X_DATA_SIZE;
const long NumPixelsAttrib::X_DATA_SIZE;
const long RegionOfInterestAttrib::X_DATA_SIZE;
const long AnalysisRegionAttrib::X_DATA_SIZE;
const long GaussFitAttrib::X_DATA_SIZE;
const long FWHMAttrib::X_DATA_SIZE;
const long GaussReferenceAttrib::X_DATA_SIZE;
const long DataReferenceAttrib::X_DATA_SIZE;
const long IntensityCalibrationAttrib::X_DATA_SIZE;
const long NoiseProfileAttrib::X_DATA_SIZE;
const long WaveAttrib::X_DATA_SIZE;
const long LambdaAttrib::X_DATA_SIZE;
//--------------------------------------------------------
/**
 * method : 		CCDSpectrometerClass::CCDSpectrometerClass(std::string &s)
 * description : 	constructor for the CCDSpectrometerClass
 *
 * @param s	The class name
 */
//--------------------------------------------------------
CCDSpectrometerClass::CCDSpectrometerClass(std::string &s):Tango::DeviceClass(s)
{
	TANGO_LOG_INFO << "Entering CCDSpectrometerClass constructor" << std::endl;
	set_default_property();
	write_class_property();

	/*----- PROTECTED REGION ID(CCDSpectrometerClass::constructor) ENABLED START -----*/
	/* clang-format on */
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CCDSpectrometerClass::constructor

	TANGO_LOG_INFO << "Leaving CCDSpectrometerClass constructor" << std::endl;
}

//--------------------------------------------------------
/**
 * method : 		CCDSpectrometerClass::~CCDSpectrometerClass()
 * description : 	destructor for the CCDSpectrometerClass
 */
//--------------------------------------------------------
CCDSpectrometerClass::~CCDSpectrometerClass()
{
	/*----- PROTECTED REGION ID(CCDSpectrometerClass::destructor) ENABLED START -----*/
	/* clang-format on */
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CCDSpectrometerClass::destructor

	_instance = NULL;
}


//--------------------------------------------------------
/**
 * method : 		CCDSpectrometerClass::init
 * description : 	Create the object if not already done.
 *                  Otherwise, just return a pointer to the object
 *
 * @param	name	The class name
 */
//--------------------------------------------------------
CCDSpectrometerClass *CCDSpectrometerClass::init(const char *name)
{
	if (_instance == NULL)
	{
		try
		{
			std::string s(name);
			_instance = new CCDSpectrometerClass(s);
		}
		catch (std::bad_alloc &)
		{
			throw;
		}
	}
	return _instance;
}

//--------------------------------------------------------
/**
 * method : 		CCDSpectrometerClass::instance
 * description : 	Check if object already created,
 *                  and return a pointer to the object
 */
//--------------------------------------------------------
CCDSpectrometerClass *CCDSpectrometerClass::instance()
{
	if (_instance == NULL)
	{
		std::cerr << "Class is not initialized !!" << std::endl;
		exit(-1);
	}
	return _instance;
}



//===================================================================
//	Command execution method calls
//===================================================================
//--------------------------------------------------------
/**
 * method : 		StartClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *StartClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "StartClass::execute(): arrived" << std::endl;
	((static_cast<CCDSpectrometer *>(device))->start());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		StopClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *StopClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "StopClass::execute(): arrived" << std::endl;
	((static_cast<CCDSpectrometer *>(device))->stop());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		ResetClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *ResetClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "ResetClass::execute(): arrived" << std::endl;
	((static_cast<CCDSpectrometer *>(device))->reset());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		StartSingleClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *StartSingleClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "StartSingleClass::execute(): arrived" << std::endl;
	((static_cast<CCDSpectrometer *>(device))->start_single());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		StartContinuousClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *StartContinuousClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "StartContinuousClass::execute(): arrived" << std::endl;
	((static_cast<CCDSpectrometer *>(device))->start_continuous());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		GetLastDataClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *GetLastDataClass::execute(Tango::DeviceImpl *device, const CORBA::Any &in_any)
{
	TANGO_LOG_INFO << "GetLastDataClass::execute(): arrived" << std::endl;
	Tango::DevLong argin;
	extract(in_any, argin);
	return insert((static_cast<CCDSpectrometer *>(device))->get_last_data(argin));
}

//--------------------------------------------------------
/**
 * method : 		AcquireNoiseClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *AcquireNoiseClass::execute(Tango::DeviceImpl *device, const CORBA::Any &in_any)
{
	TANGO_LOG_INFO << "AcquireNoiseClass::execute(): arrived" << std::endl;
	Tango::DevLong argin;
	extract(in_any, argin);
	((static_cast<CCDSpectrometer *>(device))->acquire_noise(argin));
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		SaveReferenceDataClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *SaveReferenceDataClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "SaveReferenceDataClass::execute(): arrived" << std::endl;
	((static_cast<CCDSpectrometer *>(device))->save_reference_data());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		GetDataClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *GetDataClass::execute(Tango::DeviceImpl *device, const CORBA::Any &in_any)
{
	TANGO_LOG_INFO << "GetDataClass::execute(): arrived" << std::endl;
	const Tango::DevVarLongArray *argin;
	extract(in_any, argin);
	return insert((static_cast<CCDSpectrometer *>(device))->get_data(argin));
}

//--------------------------------------------------------
/**
 * method : 		GetGaussFitCenterClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *GetGaussFitCenterClass::execute(Tango::DeviceImpl *device, const CORBA::Any &in_any)
{
	TANGO_LOG_INFO << "GetGaussFitCenterClass::execute(): arrived" << std::endl;
	const Tango::DevVarLongArray *argin;
	extract(in_any, argin);
	return insert((static_cast<CCDSpectrometer *>(device))->get_gauss_fit_center(argin));
}

//--------------------------------------------------------
/**
 * method : 		GetGaussFitWidthClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *GetGaussFitWidthClass::execute(Tango::DeviceImpl *device, const CORBA::Any &in_any)
{
	TANGO_LOG_INFO << "GetGaussFitWidthClass::execute(): arrived" << std::endl;
	const Tango::DevVarLongArray *argin;
	extract(in_any, argin);
	return insert((static_cast<CCDSpectrometer *>(device))->get_gauss_fit_width(argin));
}

//--------------------------------------------------------
/**
 * method : 		GetFWHMCenterClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *GetFWHMCenterClass::execute(Tango::DeviceImpl *device, const CORBA::Any &in_any)
{
	TANGO_LOG_INFO << "GetFWHMCenterClass::execute(): arrived" << std::endl;
	const Tango::DevVarLongArray *argin;
	extract(in_any, argin);
	return insert((static_cast<CCDSpectrometer *>(device))->get_fwhmcenter(argin));
}

//--------------------------------------------------------
/**
 * method : 		GetFWHMWidthClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *GetFWHMWidthClass::execute(Tango::DeviceImpl *device, const CORBA::Any &in_any)
{
	TANGO_LOG_INFO << "GetFWHMWidthClass::execute(): arrived" << std::endl;
	const Tango::DevVarLongArray *argin;
	extract(in_any, argin);
	return insert((static_cast<CCDSpectrometer *>(device))->get_fwhmwidth(argin));
}

//--------------------------------------------------------
/**
 * method : 		ExportWavelengthClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *ExportWavelengthClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "ExportWavelengthClass::execute(): arrived" << std::endl;
	((static_cast<CCDSpectrometer *>(device))->export_wavelength());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		GetAreaClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *GetAreaClass::execute(Tango::DeviceImpl *device, const CORBA::Any &in_any)
{
	TANGO_LOG_INFO << "GetAreaClass::execute(): arrived" << std::endl;
	const Tango::DevVarLongArray *argin;
	extract(in_any, argin);
	return insert((static_cast<CCDSpectrometer *>(device))->get_area(argin));
}

//--------------------------------------------------------
/**
 * method : 		GetWaveClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *GetWaveClass::execute(Tango::DeviceImpl *device, const CORBA::Any &in_any)
{
	TANGO_LOG_INFO << "GetWaveClass::execute(): arrived" << std::endl;
	const Tango::DevVarLongArray *argin;
	extract(in_any, argin);
	return insert((static_cast<CCDSpectrometer *>(device))->get_wave(argin));
}

//--------------------------------------------------------
/**
 * method : 		GetLambdaClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *GetLambdaClass::execute(Tango::DeviceImpl *device, const CORBA::Any &in_any)
{
	TANGO_LOG_INFO << "GetLambdaClass::execute(): arrived" << std::endl;
	const Tango::DevVarLongArray *argin;
	extract(in_any, argin);
	return insert((static_cast<CCDSpectrometer *>(device))->get_lambda(argin));
}


//===================================================================
//	Properties management
//===================================================================
//--------------------------------------------------------
/**
 *	Method     : CCDSpectrometerClass::get_class_property()
 *	Description: Get the class property for specified name.
 */
//--------------------------------------------------------
Tango::DbDatum CCDSpectrometerClass::get_class_property(std::string &prop_name)
{
	for (unsigned int i=0 ; i<cl_prop.size() ; i++)
		if (cl_prop[i].name == prop_name)
			return cl_prop[i];
	//	if not found, returns  an empty DbDatum
	return Tango::DbDatum(prop_name);
}

//--------------------------------------------------------
/**
 *	Method     : CCDSpectrometerClass::get_default_device_property()
 *	Description: Return the default value for device property.
 */
//--------------------------------------------------------
Tango::DbDatum CCDSpectrometerClass::get_default_device_property(std::string &prop_name)
{
	for (unsigned int i=0 ; i<dev_def_prop.size() ; i++)
		if (dev_def_prop[i].name == prop_name)
			return dev_def_prop[i];
	//	if not found, return  an empty DbDatum
	return Tango::DbDatum(prop_name);
}

//--------------------------------------------------------
/**
 *	Method     : CCDSpectrometerClass::get_default_class_property()
 *	Description: Return the default value for class property.
 */
//--------------------------------------------------------
Tango::DbDatum CCDSpectrometerClass::get_default_class_property(std::string &prop_name)
{
	for (unsigned int i=0 ; i<cl_def_prop.size() ; i++)
		if (cl_def_prop[i].name == prop_name)
			return cl_def_prop[i];
	//	if not found, return  an empty DbDatum
	return Tango::DbDatum(prop_name);
}


//--------------------------------------------------------
/**
 *	Method     : CCDSpectrometerClass::set_default_property()
 *	Description: Set default property (class and device) for wizard.
 *                For each property, add to wizard property name and description.
 *                If default value has been set, add it to wizard property and
 *                store it in a DbDatum.
 */
//--------------------------------------------------------
void CCDSpectrometerClass::set_default_property()
{
	std::string	prop_name;
	std::string	prop_desc;
	std::string	prop_def;
	std::vector<std::string>	vect_data;

	//	Set Default Class Properties

	//	Set Default device Properties
	prop_name = "DeviceName";
	prop_desc = "";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "DefaultGaussRef";
	prop_desc = "Default Gauss Reference";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "DataReferenceFile";
	prop_desc = "Default Data Reference File";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "DefaultStartRegion";
	prop_desc = "Beginning of the region of interest";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "DefaultEndRegion";
	prop_desc = "End of the region of interest";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "PreferFWHM";
	prop_desc = "Choose FWHM instead og GaussianFit if both enabled";
	prop_def  = "false";
	vect_data.clear();
	vect_data.push_back("false");
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "AutoExposureRange";
	prop_desc = "AutoExposure Range";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "AutoExposureStep";
	prop_desc = "AutoExposure Step";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "AutoExposureAmpRange";
	prop_desc = "AutoExposure Amp Range";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "IntensityCalibrationFile";
	prop_desc = "";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "NoiseProfileFile";
	prop_desc = "";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "WavelengthAttr";
	prop_desc = "device name\nattribute name";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "AreaInAnalysisRegion";
	prop_desc = "use AnalysisRegion for area instead of AreaRegion";
	prop_def  = "false";
	vect_data.clear();
	vect_data.push_back("false");
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "WidthCorrection";
	prop_desc = "widht_in^2 + widthcorrection^2 = widht_out^2";
	prop_def  = "0";
	vect_data.clear();
	vect_data.push_back("0");
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "NFramesBuffered";
	prop_desc = "number of frames buffered when in live mode";
	prop_def  = "1000";
	vect_data.clear();
	vect_data.push_back("1000");
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "AnalysisRegion";
	prop_desc = "AnalysisRegion";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "RegionOfInterest";
	prop_desc = "RegionOfInterest";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "ExposureTime";
	prop_desc = "ExposureTime";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "SyncMode";
	prop_desc = "ExposureTime";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "AutoStart";
	prop_desc = "AutoStart";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "NumPixel";
	prop_desc = "Number of pixel of the CCD";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "CameraLabel";
	prop_desc = "Camera label or IP";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
}

//--------------------------------------------------------
/**
 *	Method     : CCDSpectrometerClass::write_class_property()
 *	Description: Set class description fields as property in database
 */
//--------------------------------------------------------
void CCDSpectrometerClass::write_class_property()
{
	//	First time, check if database used
	if (Tango::Util::_UseDb == false)
		return;

	Tango::DbData	data;
	std::string	classname = get_name();
	std::string	header;

	//	Put title
	Tango::DbDatum	title("ProjectTitle");
	std::string	str_title("CCDSpectrometer");
	title << str_title;
	data.push_back(title);

	//	Put Description
	Tango::DbDatum	description("Description");
	std::vector<std::string>	str_desc;
	str_desc.push_back("Laser Spectrometer based on Basler CCD Tango Device");
	description << str_desc;
	data.push_back(description);

	//  Put inheritance
	Tango::DbDatum	inher_datum("InheritedFrom");
	std::vector<std::string> inheritance;
	inheritance.push_back("TANGO_BASE_CLASS");
	inher_datum << inheritance;
	data.push_back(inher_datum);

	//	Call database and and values
	get_db_class()->put_property(data);
}

//===================================================================
//	Factory methods
//===================================================================

//--------------------------------------------------------
/**
 *	Method     : CCDSpectrometerClass::device_factory()
 *	Description: Create the device object(s)
 *                and store them in the device list
 */
//--------------------------------------------------------
void CCDSpectrometerClass::device_factory(const Tango::DevVarStringArray *devlist_ptr)
{
	/*----- PROTECTED REGION ID(CCDSpectrometerClass::device_factory_before) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CCDSpectrometerClass::device_factory_before

	//	Create devices and add it into the device list
	for (unsigned long i=0 ; i<devlist_ptr->length() ; i++)
	{
		TANGO_LOG_DEBUG << "Device name : " << (*devlist_ptr)[i].in() << std::endl;
		device_list.push_back(new CCDSpectrometer(this, (*devlist_ptr)[i]));
	}

	//	Manage dynamic attributes if any
	erase_dynamic_attributes(devlist_ptr, get_class_attr()->get_attr_list());

	//	Export devices to the outside world
	for (unsigned long i=1 ; i<=devlist_ptr->length() ; i++)
	{
		//	Add dynamic attributes if any
		CCDSpectrometer *dev = static_cast<CCDSpectrometer *>(device_list[device_list.size()-i]);
		dev->add_dynamic_attributes();

		//	Check before if database used.
		if ((Tango::Util::_UseDb == true) && (Tango::Util::_FileDb == false))
			export_device(dev);
		else
			export_device(dev, dev->get_name().c_str());
	}

	/*----- PROTECTED REGION ID(CCDSpectrometerClass::device_factory_after) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CCDSpectrometerClass::device_factory_after
}
//--------------------------------------------------------
/**
 *	Method     : CCDSpectrometerClass::attribute_factory()
 *	Description: Create the attribute object(s)
 *                and store them in the attribute list
 */
//--------------------------------------------------------
void CCDSpectrometerClass::attribute_factory(std::vector<Tango::Attr *> &att_list)
{
	/*----- PROTECTED REGION ID(CCDSpectrometerClass::attribute_factory_before) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CCDSpectrometerClass::attribute_factory_before
	//	Attribute : SyncMode
	SyncModeAttrib	*syncmode = new SyncModeAttrib();
	Tango::UserDefaultAttrProp	syncmode_prop;
	syncmode_prop.set_description("0=SW\n1=HW Sync input\n2=HW external trigger");
	//	label	not set for SyncMode
	//	unit	not set for SyncMode
	//	standard_unit	not set for SyncMode
	//	display_unit	not set for SyncMode
	//	format	not set for SyncMode
	//	max_value	not set for SyncMode
	//	min_value	not set for SyncMode
	//	max_alarm	not set for SyncMode
	//	min_alarm	not set for SyncMode
	//	max_warning	not set for SyncMode
	//	min_warning	not set for SyncMode
	//	delta_t	not set for SyncMode
	//	delta_val	not set for SyncMode
	syncmode->set_default_properties(syncmode_prop);
	//	Not Polled
	syncmode->set_disp_level(Tango::OPERATOR);
	syncmode->set_memorized();
	syncmode->set_memorized_init(false);
	att_list.push_back(syncmode);

	//	Attribute : ExposureTime
	ExposureTimeAttrib	*exposuretime = new ExposureTimeAttrib();
	Tango::UserDefaultAttrProp	exposuretime_prop;
	exposuretime_prop.set_description("Exposure time in milliseconds");
	//	label	not set for ExposureTime
	exposuretime_prop.set_unit("ms");
	exposuretime_prop.set_standard_unit("1");
	exposuretime_prop.set_display_unit("ms");
	//	format	not set for ExposureTime
	//	max_value	not set for ExposureTime
	//	min_value	not set for ExposureTime
	//	max_alarm	not set for ExposureTime
	//	min_alarm	not set for ExposureTime
	//	max_warning	not set for ExposureTime
	//	min_warning	not set for ExposureTime
	//	delta_t	not set for ExposureTime
	//	delta_val	not set for ExposureTime
	exposuretime->set_default_properties(exposuretime_prop);
	//	Not Polled
	exposuretime->set_disp_level(Tango::OPERATOR);
	exposuretime->set_memorized();
	exposuretime->set_memorized_init(false);
	att_list.push_back(exposuretime);

	//	Attribute : NoiseSubtraction
	NoiseSubtractionAttrib	*noisesubtraction = new NoiseSubtractionAttrib();
	Tango::UserDefaultAttrProp	noisesubtraction_prop;
	noisesubtraction_prop.set_description("Enable/Disable subtraction of background noise profile \nacquired with the AcquireNoise command");
	//	label	not set for NoiseSubtraction
	//	unit	not set for NoiseSubtraction
	//	standard_unit	not set for NoiseSubtraction
	//	display_unit	not set for NoiseSubtraction
	//	format	not set for NoiseSubtraction
	//	max_value	not set for NoiseSubtraction
	//	min_value	not set for NoiseSubtraction
	//	max_alarm	not set for NoiseSubtraction
	//	min_alarm	not set for NoiseSubtraction
	//	max_warning	not set for NoiseSubtraction
	//	min_warning	not set for NoiseSubtraction
	//	delta_t	not set for NoiseSubtraction
	//	delta_val	not set for NoiseSubtraction
	noisesubtraction->set_default_properties(noisesubtraction_prop);
	//	Not Polled
	noisesubtraction->set_disp_level(Tango::OPERATOR);
	noisesubtraction->set_memorized();
	noisesubtraction->set_memorized_init(true);
	att_list.push_back(noisesubtraction);

	//	Attribute : GaussFitEnable
	GaussFitEnableAttrib	*gaussfitenable = new GaussFitEnableAttrib();
	Tango::UserDefaultAttrProp	gaussfitenable_prop;
	//	description	not set for GaussFitEnable
	//	label	not set for GaussFitEnable
	//	unit	not set for GaussFitEnable
	//	standard_unit	not set for GaussFitEnable
	//	display_unit	not set for GaussFitEnable
	//	format	not set for GaussFitEnable
	//	max_value	not set for GaussFitEnable
	//	min_value	not set for GaussFitEnable
	//	max_alarm	not set for GaussFitEnable
	//	min_alarm	not set for GaussFitEnable
	//	max_warning	not set for GaussFitEnable
	//	min_warning	not set for GaussFitEnable
	//	delta_t	not set for GaussFitEnable
	//	delta_val	not set for GaussFitEnable
	gaussfitenable->set_default_properties(gaussfitenable_prop);
	//	Not Polled
	gaussfitenable->set_disp_level(Tango::OPERATOR);
	gaussfitenable->set_memorized();
	gaussfitenable->set_memorized_init(true);
	att_list.push_back(gaussfitenable);

	//	Attribute : FWHMEnable
	FWHMEnableAttrib	*fwhmenable = new FWHMEnableAttrib();
	Tango::UserDefaultAttrProp	fwhmenable_prop;
	//	description	not set for FWHMEnable
	//	label	not set for FWHMEnable
	//	unit	not set for FWHMEnable
	//	standard_unit	not set for FWHMEnable
	//	display_unit	not set for FWHMEnable
	//	format	not set for FWHMEnable
	//	max_value	not set for FWHMEnable
	//	min_value	not set for FWHMEnable
	//	max_alarm	not set for FWHMEnable
	//	min_alarm	not set for FWHMEnable
	//	max_warning	not set for FWHMEnable
	//	min_warning	not set for FWHMEnable
	//	delta_t	not set for FWHMEnable
	//	delta_val	not set for FWHMEnable
	fwhmenable->set_default_properties(fwhmenable_prop);
	//	Not Polled
	fwhmenable->set_disp_level(Tango::OPERATOR);
	fwhmenable->set_memorized();
	fwhmenable->set_memorized_init(true);
	att_list.push_back(fwhmenable);

	//	Attribute : GaussFitCenter
	GaussFitCenterAttrib	*gaussfitcenter = new GaussFitCenterAttrib();
	Tango::UserDefaultAttrProp	gaussfitcenter_prop;
	gaussfitcenter_prop.set_description("wavelength of peak of Gaussian Fit");
	//	label	not set for GaussFitCenter
	//	unit	not set for GaussFitCenter
	//	standard_unit	not set for GaussFitCenter
	//	display_unit	not set for GaussFitCenter
	//	format	not set for GaussFitCenter
	//	max_value	not set for GaussFitCenter
	//	min_value	not set for GaussFitCenter
	//	max_alarm	not set for GaussFitCenter
	//	min_alarm	not set for GaussFitCenter
	//	max_warning	not set for GaussFitCenter
	//	min_warning	not set for GaussFitCenter
	//	delta_t	not set for GaussFitCenter
	//	delta_val	not set for GaussFitCenter
	gaussfitcenter->set_default_properties(gaussfitcenter_prop);
	//	Not Polled
	gaussfitcenter->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(gaussfitcenter);

	//	Attribute : GaussFitWidth
	GaussFitWidthAttrib	*gaussfitwidth = new GaussFitWidthAttrib();
	Tango::UserDefaultAttrProp	gaussfitwidth_prop;
	gaussfitwidth_prop.set_description("width of Gaussian Fit");
	//	label	not set for GaussFitWidth
	//	unit	not set for GaussFitWidth
	//	standard_unit	not set for GaussFitWidth
	//	display_unit	not set for GaussFitWidth
	//	format	not set for GaussFitWidth
	//	max_value	not set for GaussFitWidth
	//	min_value	not set for GaussFitWidth
	//	max_alarm	not set for GaussFitWidth
	//	min_alarm	not set for GaussFitWidth
	//	max_warning	not set for GaussFitWidth
	//	min_warning	not set for GaussFitWidth
	//	delta_t	not set for GaussFitWidth
	//	delta_val	not set for GaussFitWidth
	gaussfitwidth->set_default_properties(gaussfitwidth_prop);
	//	Not Polled
	gaussfitwidth->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(gaussfitwidth);

	//	Attribute : FWHMCenter
	FWHMCenterAttrib	*fwhmcenter = new FWHMCenterAttrib();
	Tango::UserDefaultAttrProp	fwhmcenter_prop;
	fwhmcenter_prop.set_description("wavelength of peak of FWHM");
	//	label	not set for FWHMCenter
	//	unit	not set for FWHMCenter
	//	standard_unit	not set for FWHMCenter
	//	display_unit	not set for FWHMCenter
	//	format	not set for FWHMCenter
	//	max_value	not set for FWHMCenter
	//	min_value	not set for FWHMCenter
	//	max_alarm	not set for FWHMCenter
	//	min_alarm	not set for FWHMCenter
	//	max_warning	not set for FWHMCenter
	//	min_warning	not set for FWHMCenter
	//	delta_t	not set for FWHMCenter
	//	delta_val	not set for FWHMCenter
	fwhmcenter->set_default_properties(fwhmcenter_prop);
	//	Not Polled
	fwhmcenter->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(fwhmcenter);

	//	Attribute : FWHMWidth
	FWHMWidthAttrib	*fwhmwidth = new FWHMWidthAttrib();
	Tango::UserDefaultAttrProp	fwhmwidth_prop;
	fwhmwidth_prop.set_description("width of FWHM");
	//	label	not set for FWHMWidth
	//	unit	not set for FWHMWidth
	//	standard_unit	not set for FWHMWidth
	//	display_unit	not set for FWHMWidth
	//	format	not set for FWHMWidth
	//	max_value	not set for FWHMWidth
	//	min_value	not set for FWHMWidth
	//	max_alarm	not set for FWHMWidth
	//	min_alarm	not set for FWHMWidth
	//	max_warning	not set for FWHMWidth
	//	min_warning	not set for FWHMWidth
	//	delta_t	not set for FWHMWidth
	//	delta_val	not set for FWHMWidth
	fwhmwidth->set_default_properties(fwhmwidth_prop);
	//	Not Polled
	fwhmwidth->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(fwhmwidth);

	//	Attribute : Center
	CenterAttrib	*center = new CenterAttrib();
	Tango::UserDefaultAttrProp	center_prop;
	center_prop.set_description("wavelength of peak of FWHM or of Gaussian Fit");
	//	label	not set for Center
	//	unit	not set for Center
	//	standard_unit	not set for Center
	//	display_unit	not set for Center
	//	format	not set for Center
	//	max_value	not set for Center
	//	min_value	not set for Center
	//	max_alarm	not set for Center
	//	min_alarm	not set for Center
	//	max_warning	not set for Center
	//	min_warning	not set for Center
	//	delta_t	not set for Center
	//	delta_val	not set for Center
	center->set_default_properties(center_prop);
	//	Not Polled
	center->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(center);

	//	Attribute : Width
	WidthAttrib	*width = new WidthAttrib();
	Tango::UserDefaultAttrProp	width_prop;
	width_prop.set_description("width of FWHM or of Gaussian Fit");
	//	label	not set for Width
	//	unit	not set for Width
	//	standard_unit	not set for Width
	//	display_unit	not set for Width
	//	format	not set for Width
	//	max_value	not set for Width
	//	min_value	not set for Width
	//	max_alarm	not set for Width
	//	min_alarm	not set for Width
	//	max_warning	not set for Width
	//	min_warning	not set for Width
	//	delta_t	not set for Width
	//	delta_val	not set for Width
	width->set_default_properties(width_prop);
	//	Not Polled
	width->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(width);

	//	Attribute : FitType
	FitTypeAttrib	*fittype = new FitTypeAttrib();
	Tango::UserDefaultAttrProp	fittype_prop;
	fittype_prop.set_description("1=Gaussian 2=Gaussian Asymm 3=Confiteor");
	//	label	not set for FitType
	//	unit	not set for FitType
	//	standard_unit	not set for FitType
	//	display_unit	not set for FitType
	//	format	not set for FitType
	fittype_prop.set_max_value("3");
	fittype_prop.set_min_value("1");
	//	max_alarm	not set for FitType
	//	min_alarm	not set for FitType
	//	max_warning	not set for FitType
	//	min_warning	not set for FitType
	//	delta_t	not set for FitType
	//	delta_val	not set for FitType
	fittype->set_default_properties(fittype_prop);
	//	Not Polled
	fittype->set_disp_level(Tango::OPERATOR);
	fittype->set_memorized();
	fittype->set_memorized_init(true);
	att_list.push_back(fittype);

	//	Attribute : FWHMAvg
	FWHMAvgAttrib	*fwhmavg = new FWHMAvgAttrib();
	Tango::UserDefaultAttrProp	fwhmavg_prop;
	//	description	not set for FWHMAvg
	//	label	not set for FWHMAvg
	//	unit	not set for FWHMAvg
	//	standard_unit	not set for FWHMAvg
	//	display_unit	not set for FWHMAvg
	//	format	not set for FWHMAvg
	//	max_value	not set for FWHMAvg
	//	min_value	not set for FWHMAvg
	//	max_alarm	not set for FWHMAvg
	//	min_alarm	not set for FWHMAvg
	//	max_warning	not set for FWHMAvg
	//	min_warning	not set for FWHMAvg
	//	delta_t	not set for FWHMAvg
	//	delta_val	not set for FWHMAvg
	fwhmavg->set_default_properties(fwhmavg_prop);
	//	Not Polled
	fwhmavg->set_disp_level(Tango::OPERATOR);
	fwhmavg->set_memorized();
	fwhmavg->set_memorized_init(true);
	att_list.push_back(fwhmavg);

	//	Attribute : NoiseLevel
	NoiseLevelAttrib	*noiselevel = new NoiseLevelAttrib();
	Tango::UserDefaultAttrProp	noiselevel_prop;
	//	description	not set for NoiseLevel
	//	label	not set for NoiseLevel
	//	unit	not set for NoiseLevel
	//	standard_unit	not set for NoiseLevel
	//	display_unit	not set for NoiseLevel
	//	format	not set for NoiseLevel
	//	max_value	not set for NoiseLevel
	//	min_value	not set for NoiseLevel
	//	max_alarm	not set for NoiseLevel
	//	min_alarm	not set for NoiseLevel
	//	max_warning	not set for NoiseLevel
	//	min_warning	not set for NoiseLevel
	//	delta_t	not set for NoiseLevel
	//	delta_val	not set for NoiseLevel
	noiselevel->set_default_properties(noiselevel_prop);
	//	Not Polled
	noiselevel->set_disp_level(Tango::OPERATOR);
	noiselevel->set_memorized();
	noiselevel->set_memorized_init(true);
	att_list.push_back(noiselevel);

	//	Attribute : DataReferenceFile
	DataReferenceFileAttrib	*datareferencefile = new DataReferenceFileAttrib();
	Tango::UserDefaultAttrProp	datareferencefile_prop;
	//	description	not set for DataReferenceFile
	//	label	not set for DataReferenceFile
	//	unit	not set for DataReferenceFile
	//	standard_unit	not set for DataReferenceFile
	//	display_unit	not set for DataReferenceFile
	//	format	not set for DataReferenceFile
	//	max_value	not set for DataReferenceFile
	//	min_value	not set for DataReferenceFile
	//	max_alarm	not set for DataReferenceFile
	//	min_alarm	not set for DataReferenceFile
	//	max_warning	not set for DataReferenceFile
	//	min_warning	not set for DataReferenceFile
	//	delta_t	not set for DataReferenceFile
	//	delta_val	not set for DataReferenceFile
	datareferencefile->set_default_properties(datareferencefile_prop);
	//	Not Polled
	datareferencefile->set_disp_level(Tango::OPERATOR);
	datareferencefile->set_memorized();
	datareferencefile->set_memorized_init(false);
	att_list.push_back(datareferencefile);

	//	Attribute : AutoExposure
	AutoExposureAttrib	*autoexposure = new AutoExposureAttrib();
	Tango::UserDefaultAttrProp	autoexposure_prop;
	//	description	not set for AutoExposure
	//	label	not set for AutoExposure
	//	unit	not set for AutoExposure
	//	standard_unit	not set for AutoExposure
	//	display_unit	not set for AutoExposure
	//	format	not set for AutoExposure
	//	max_value	not set for AutoExposure
	//	min_value	not set for AutoExposure
	//	max_alarm	not set for AutoExposure
	//	min_alarm	not set for AutoExposure
	//	max_warning	not set for AutoExposure
	//	min_warning	not set for AutoExposure
	//	delta_t	not set for AutoExposure
	//	delta_val	not set for AutoExposure
	autoexposure->set_default_properties(autoexposure_prop);
	//	Not Polled
	autoexposure->set_disp_level(Tango::OPERATOR);
	autoexposure->set_memorized();
	autoexposure->set_memorized_init(true);
	att_list.push_back(autoexposure);

	//	Attribute : RtMode
	RtModeAttrib	*rtmode = new RtModeAttrib();
	Tango::UserDefaultAttrProp	rtmode_prop;
	//	description	not set for RtMode
	//	label	not set for RtMode
	//	unit	not set for RtMode
	//	standard_unit	not set for RtMode
	//	display_unit	not set for RtMode
	//	format	not set for RtMode
	//	max_value	not set for RtMode
	//	min_value	not set for RtMode
	//	max_alarm	not set for RtMode
	//	min_alarm	not set for RtMode
	//	max_warning	not set for RtMode
	//	min_warning	not set for RtMode
	//	delta_t	not set for RtMode
	//	delta_val	not set for RtMode
	rtmode->set_default_properties(rtmode_prop);
	//	Not Polled
	rtmode->set_disp_level(Tango::OPERATOR);
	rtmode->set_memorized();
	rtmode->set_memorized_init(true);
	att_list.push_back(rtmode);

	//	Attribute : IntensityCalibrationEnable
	IntensityCalibrationEnableAttrib	*intensitycalibrationenable = new IntensityCalibrationEnableAttrib();
	Tango::UserDefaultAttrProp	intensitycalibrationenable_prop;
	//	description	not set for IntensityCalibrationEnable
	//	label	not set for IntensityCalibrationEnable
	//	unit	not set for IntensityCalibrationEnable
	//	standard_unit	not set for IntensityCalibrationEnable
	//	display_unit	not set for IntensityCalibrationEnable
	//	format	not set for IntensityCalibrationEnable
	//	max_value	not set for IntensityCalibrationEnable
	//	min_value	not set for IntensityCalibrationEnable
	//	max_alarm	not set for IntensityCalibrationEnable
	//	min_alarm	not set for IntensityCalibrationEnable
	//	max_warning	not set for IntensityCalibrationEnable
	//	min_warning	not set for IntensityCalibrationEnable
	//	delta_t	not set for IntensityCalibrationEnable
	//	delta_val	not set for IntensityCalibrationEnable
	intensitycalibrationenable->set_default_properties(intensitycalibrationenable_prop);
	//	Not Polled
	intensitycalibrationenable->set_disp_level(Tango::OPERATOR);
	intensitycalibrationenable->set_memorized();
	intensitycalibrationenable->set_memorized_init(true);
	att_list.push_back(intensitycalibrationenable);

	//	Attribute : Area
	AreaAttrib	*area = new AreaAttrib();
	Tango::UserDefaultAttrProp	area_prop;
	//	description	not set for Area
	//	label	not set for Area
	//	unit	not set for Area
	//	standard_unit	not set for Area
	//	display_unit	not set for Area
	//	format	not set for Area
	//	max_value	not set for Area
	//	min_value	not set for Area
	//	max_alarm	not set for Area
	//	min_alarm	not set for Area
	//	max_warning	not set for Area
	//	min_warning	not set for Area
	//	delta_t	not set for Area
	//	delta_val	not set for Area
	area->set_default_properties(area_prop);
	//	Not Polled
	area->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(area);

	//	Attribute : AreaRegion
	AreaRegionAttrib	*arearegion = new AreaRegionAttrib();
	Tango::UserDefaultAttrProp	arearegion_prop;
	//	description	not set for AreaRegion
	//	label	not set for AreaRegion
	//	unit	not set for AreaRegion
	//	standard_unit	not set for AreaRegion
	//	display_unit	not set for AreaRegion
	//	format	not set for AreaRegion
	//	max_value	not set for AreaRegion
	//	min_value	not set for AreaRegion
	//	max_alarm	not set for AreaRegion
	//	min_alarm	not set for AreaRegion
	//	max_warning	not set for AreaRegion
	//	min_warning	not set for AreaRegion
	//	delta_t	not set for AreaRegion
	//	delta_val	not set for AreaRegion
	arearegion->set_default_properties(arearegion_prop);
	//	Not Polled
	arearegion->set_disp_level(Tango::OPERATOR);
	arearegion->set_memorized();
	arearegion->set_memorized_init(true);
	att_list.push_back(arearegion);

	//	Attribute : AreaEnable
	AreaEnableAttrib	*areaenable = new AreaEnableAttrib();
	Tango::UserDefaultAttrProp	areaenable_prop;
	//	description	not set for AreaEnable
	//	label	not set for AreaEnable
	//	unit	not set for AreaEnable
	//	standard_unit	not set for AreaEnable
	//	display_unit	not set for AreaEnable
	//	format	not set for AreaEnable
	//	max_value	not set for AreaEnable
	//	min_value	not set for AreaEnable
	//	max_alarm	not set for AreaEnable
	//	min_alarm	not set for AreaEnable
	//	max_warning	not set for AreaEnable
	//	min_warning	not set for AreaEnable
	//	delta_t	not set for AreaEnable
	//	delta_val	not set for AreaEnable
	areaenable->set_default_properties(areaenable_prop);
	//	Not Polled
	areaenable->set_disp_level(Tango::OPERATOR);
	areaenable->set_memorized();
	areaenable->set_memorized_init(true);
	att_list.push_back(areaenable);

	//	Attribute : WidthCorrectionEnable
	WidthCorrectionEnableAttrib	*widthcorrectionenable = new WidthCorrectionEnableAttrib();
	Tango::UserDefaultAttrProp	widthcorrectionenable_prop;
	widthcorrectionenable_prop.set_description("enable widht correction");
	//	label	not set for WidthCorrectionEnable
	//	unit	not set for WidthCorrectionEnable
	//	standard_unit	not set for WidthCorrectionEnable
	//	display_unit	not set for WidthCorrectionEnable
	//	format	not set for WidthCorrectionEnable
	//	max_value	not set for WidthCorrectionEnable
	//	min_value	not set for WidthCorrectionEnable
	//	max_alarm	not set for WidthCorrectionEnable
	//	min_alarm	not set for WidthCorrectionEnable
	//	max_warning	not set for WidthCorrectionEnable
	//	min_warning	not set for WidthCorrectionEnable
	//	delta_t	not set for WidthCorrectionEnable
	//	delta_val	not set for WidthCorrectionEnable
	widthcorrectionenable->set_default_properties(widthcorrectionenable_prop);
	//	Not Polled
	widthcorrectionenable->set_disp_level(Tango::OPERATOR);
	widthcorrectionenable->set_memorized();
	widthcorrectionenable->set_memorized_init(true);
	att_list.push_back(widthcorrectionenable);

	//	Attribute : IncrSensitivity
	IncrSensitivityAttrib	*incrsensitivity = new IncrSensitivityAttrib();
	Tango::UserDefaultAttrProp	incrsensitivity_prop;
	incrsensitivity_prop.set_description("false: use standard sensitivity\ntrue: use increased sensitivity");
	//	label	not set for IncrSensitivity
	//	unit	not set for IncrSensitivity
	//	standard_unit	not set for IncrSensitivity
	//	display_unit	not set for IncrSensitivity
	//	format	not set for IncrSensitivity
	//	max_value	not set for IncrSensitivity
	//	min_value	not set for IncrSensitivity
	//	max_alarm	not set for IncrSensitivity
	//	min_alarm	not set for IncrSensitivity
	//	max_warning	not set for IncrSensitivity
	//	min_warning	not set for IncrSensitivity
	//	delta_t	not set for IncrSensitivity
	//	delta_val	not set for IncrSensitivity
	incrsensitivity->set_default_properties(incrsensitivity_prop);
	//	Not Polled
	incrsensitivity->set_disp_level(Tango::OPERATOR);
	incrsensitivity->set_memorized();
	incrsensitivity->set_memorized_init(true);
	att_list.push_back(incrsensitivity);

	//	Attribute : Data
	DataAttrib	*data = new DataAttrib();
	Tango::UserDefaultAttrProp	data_prop;
	//	description	not set for Data
	//	label	not set for Data
	//	unit	not set for Data
	//	standard_unit	not set for Data
	//	display_unit	not set for Data
	//	format	not set for Data
	//	max_value	not set for Data
	//	min_value	not set for Data
	//	max_alarm	not set for Data
	//	min_alarm	not set for Data
	//	max_warning	not set for Data
	//	min_warning	not set for Data
	//	delta_t	not set for Data
	//	delta_val	not set for Data
	data->set_default_properties(data_prop);
	//	Not Polled
	data->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(data);

	//	Attribute : NumPixels
	NumPixelsAttrib	*numpixels = new NumPixelsAttrib();
	Tango::UserDefaultAttrProp	numpixels_prop;
	numpixels_prop.set_description("long[0]: number of pixels horizontally (number of CCD-array columns)\nlong[1]: number of pixels vertically (number of CCD-array lines)");
	//	label	not set for NumPixels
	//	unit	not set for NumPixels
	//	standard_unit	not set for NumPixels
	//	display_unit	not set for NumPixels
	//	format	not set for NumPixels
	//	max_value	not set for NumPixels
	//	min_value	not set for NumPixels
	//	max_alarm	not set for NumPixels
	//	min_alarm	not set for NumPixels
	//	max_warning	not set for NumPixels
	//	min_warning	not set for NumPixels
	//	delta_t	not set for NumPixels
	//	delta_val	not set for NumPixels
	numpixels->set_default_properties(numpixels_prop);
	//	Not Polled
	numpixels->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(numpixels);

	//	Attribute : RegionOfInterest
	RegionOfInterestAttrib	*regionofinterest = new RegionOfInterestAttrib();
	Tango::UserDefaultAttrProp	regionofinterest_prop;
	regionofinterest_prop.set_description("Region of interest returned by the Data attribute\ndouble[0]=start of the region\ndouble[1]=end of the region\n\nValues are in nm according with the conversion factor in x_values\nData attribute property");
	//	label	not set for RegionOfInterest
	//	unit	not set for RegionOfInterest
	//	standard_unit	not set for RegionOfInterest
	//	display_unit	not set for RegionOfInterest
	//	format	not set for RegionOfInterest
	//	max_value	not set for RegionOfInterest
	//	min_value	not set for RegionOfInterest
	//	max_alarm	not set for RegionOfInterest
	//	min_alarm	not set for RegionOfInterest
	//	max_warning	not set for RegionOfInterest
	//	min_warning	not set for RegionOfInterest
	//	delta_t	not set for RegionOfInterest
	//	delta_val	not set for RegionOfInterest
	regionofinterest->set_default_properties(regionofinterest_prop);
	//	Not Polled
	regionofinterest->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(regionofinterest);

	//	Attribute : AnalysisRegion
	AnalysisRegionAttrib	*analysisregion = new AnalysisRegionAttrib();
	Tango::UserDefaultAttrProp	analysisregion_prop;
	analysisregion_prop.set_description("Region used for gaussian and fwhm analysis\ndouble[0]=start of the region\ndouble[1]=end of the region\n\nValues are in nm according with the conversion factor in x_values\nData attribute property");
	//	label	not set for AnalysisRegion
	//	unit	not set for AnalysisRegion
	//	standard_unit	not set for AnalysisRegion
	//	display_unit	not set for AnalysisRegion
	//	format	not set for AnalysisRegion
	//	max_value	not set for AnalysisRegion
	//	min_value	not set for AnalysisRegion
	//	max_alarm	not set for AnalysisRegion
	//	min_alarm	not set for AnalysisRegion
	//	max_warning	not set for AnalysisRegion
	//	min_warning	not set for AnalysisRegion
	//	delta_t	not set for AnalysisRegion
	//	delta_val	not set for AnalysisRegion
	analysisregion->set_default_properties(analysisregion_prop);
	//	Not Polled
	analysisregion->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(analysisregion);

	//	Attribute : GaussFit
	GaussFitAttrib	*gaussfit = new GaussFitAttrib();
	Tango::UserDefaultAttrProp	gaussfit_prop;
	gaussfit_prop.set_description("double[0]= wavelength of peak of Gaussian Fit\ndouble[1]= width of Gaussian Fit");
	//	label	not set for GaussFit
	//	unit	not set for GaussFit
	//	standard_unit	not set for GaussFit
	//	display_unit	not set for GaussFit
	//	format	not set for GaussFit
	//	max_value	not set for GaussFit
	//	min_value	not set for GaussFit
	//	max_alarm	not set for GaussFit
	//	min_alarm	not set for GaussFit
	//	max_warning	not set for GaussFit
	//	min_warning	not set for GaussFit
	//	delta_t	not set for GaussFit
	//	delta_val	not set for GaussFit
	gaussfit->set_default_properties(gaussfit_prop);
	//	Not Polled
	gaussfit->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(gaussfit);

	//	Attribute : FWHM
	FWHMAttrib	*fwhm = new FWHMAttrib();
	Tango::UserDefaultAttrProp	fwhm_prop;
	fwhm_prop.set_description("double[0]= wavelength of peak of FWHM\ndouble[1]= width of FWHM");
	//	label	not set for FWHM
	//	unit	not set for FWHM
	//	standard_unit	not set for FWHM
	//	display_unit	not set for FWHM
	//	format	not set for FWHM
	//	max_value	not set for FWHM
	//	min_value	not set for FWHM
	//	max_alarm	not set for FWHM
	//	min_alarm	not set for FWHM
	//	max_warning	not set for FWHM
	//	min_warning	not set for FWHM
	//	delta_t	not set for FWHM
	//	delta_val	not set for FWHM
	fwhm->set_default_properties(fwhm_prop);
	//	Not Polled
	fwhm->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(fwhm);

	//	Attribute : GaussReference
	GaussReferenceAttrib	*gaussreference = new GaussReferenceAttrib();
	Tango::UserDefaultAttrProp	gaussreference_prop;
	gaussreference_prop.set_description("double[0]= wavelength of peak of Gaussian Fit\ndouble[1]= width of Gaussian Fit");
	//	label	not set for GaussReference
	//	unit	not set for GaussReference
	//	standard_unit	not set for GaussReference
	//	display_unit	not set for GaussReference
	//	format	not set for GaussReference
	//	max_value	not set for GaussReference
	//	min_value	not set for GaussReference
	//	max_alarm	not set for GaussReference
	//	min_alarm	not set for GaussReference
	//	max_warning	not set for GaussReference
	//	min_warning	not set for GaussReference
	//	delta_t	not set for GaussReference
	//	delta_val	not set for GaussReference
	gaussreference->set_default_properties(gaussreference_prop);
	//	Not Polled
	gaussreference->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(gaussreference);

	//	Attribute : DataReference
	DataReferenceAttrib	*datareference = new DataReferenceAttrib();
	Tango::UserDefaultAttrProp	datareference_prop;
	datareference_prop.set_description("A reference curve");
	//	label	not set for DataReference
	//	unit	not set for DataReference
	//	standard_unit	not set for DataReference
	//	display_unit	not set for DataReference
	//	format	not set for DataReference
	//	max_value	not set for DataReference
	//	min_value	not set for DataReference
	//	max_alarm	not set for DataReference
	//	min_alarm	not set for DataReference
	//	max_warning	not set for DataReference
	//	min_warning	not set for DataReference
	//	delta_t	not set for DataReference
	//	delta_val	not set for DataReference
	datareference->set_default_properties(datareference_prop);
	//	Not Polled
	datareference->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(datareference);

	//	Attribute : IntensityCalibration
	IntensityCalibrationAttrib	*intensitycalibration = new IntensityCalibrationAttrib();
	Tango::UserDefaultAttrProp	intensitycalibration_prop;
	intensitycalibration_prop.set_description("A reference curve");
	//	label	not set for IntensityCalibration
	//	unit	not set for IntensityCalibration
	//	standard_unit	not set for IntensityCalibration
	//	display_unit	not set for IntensityCalibration
	//	format	not set for IntensityCalibration
	//	max_value	not set for IntensityCalibration
	//	min_value	not set for IntensityCalibration
	//	max_alarm	not set for IntensityCalibration
	//	min_alarm	not set for IntensityCalibration
	//	max_warning	not set for IntensityCalibration
	//	min_warning	not set for IntensityCalibration
	//	delta_t	not set for IntensityCalibration
	//	delta_val	not set for IntensityCalibration
	intensitycalibration->set_default_properties(intensitycalibration_prop);
	//	Not Polled
	intensitycalibration->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(intensitycalibration);

	//	Attribute : NoiseProfile
	NoiseProfileAttrib	*noiseprofile = new NoiseProfileAttrib();
	Tango::UserDefaultAttrProp	noiseprofile_prop;
	noiseprofile_prop.set_description("A reference curve");
	//	label	not set for NoiseProfile
	//	unit	not set for NoiseProfile
	//	standard_unit	not set for NoiseProfile
	//	display_unit	not set for NoiseProfile
	//	format	not set for NoiseProfile
	//	max_value	not set for NoiseProfile
	//	min_value	not set for NoiseProfile
	//	max_alarm	not set for NoiseProfile
	//	min_alarm	not set for NoiseProfile
	//	max_warning	not set for NoiseProfile
	//	min_warning	not set for NoiseProfile
	//	delta_t	not set for NoiseProfile
	//	delta_val	not set for NoiseProfile
	noiseprofile->set_default_properties(noiseprofile_prop);
	//	Not Polled
	noiseprofile->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(noiseprofile);

	//	Attribute : Wave
	WaveAttrib	*wave = new WaveAttrib();
	Tango::UserDefaultAttrProp	wave_prop;
	//	description	not set for Wave
	//	label	not set for Wave
	//	unit	not set for Wave
	//	standard_unit	not set for Wave
	//	display_unit	not set for Wave
	//	format	not set for Wave
	//	max_value	not set for Wave
	//	min_value	not set for Wave
	//	max_alarm	not set for Wave
	//	min_alarm	not set for Wave
	//	max_warning	not set for Wave
	//	min_warning	not set for Wave
	//	delta_t	not set for Wave
	//	delta_val	not set for Wave
	wave->set_default_properties(wave_prop);
	//	Not Polled
	wave->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(wave);

	//	Attribute : Lambda
	LambdaAttrib	*lambda = new LambdaAttrib();
	Tango::UserDefaultAttrProp	lambda_prop;
	//	description	not set for Lambda
	//	label	not set for Lambda
	//	unit	not set for Lambda
	//	standard_unit	not set for Lambda
	//	display_unit	not set for Lambda
	//	format	not set for Lambda
	//	max_value	not set for Lambda
	//	min_value	not set for Lambda
	//	max_alarm	not set for Lambda
	//	min_alarm	not set for Lambda
	//	max_warning	not set for Lambda
	//	min_warning	not set for Lambda
	//	delta_t	not set for Lambda
	//	delta_val	not set for Lambda
	lambda->set_default_properties(lambda_prop);
	//	Not Polled
	lambda->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(lambda);


	//	Create a list of static attributes
	create_static_attribute_list(get_class_attr()->get_attr_list());
	/*----- PROTECTED REGION ID(CCDSpectrometerClass::attribute_factory_after) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CCDSpectrometerClass::attribute_factory_after
}
//--------------------------------------------------------
/**
 *	Method     : CCDSpectrometerClass::pipe_factory()
 *	Description: Create the pipe object(s)
 *                and store them in the pipe list
 */
//--------------------------------------------------------
void CCDSpectrometerClass::pipe_factory()
{
	/*----- PROTECTED REGION ID(CCDSpectrometerClass::pipe_factory_before) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CCDSpectrometerClass::pipe_factory_before
	/*----- PROTECTED REGION ID(CCDSpectrometerClass::pipe_factory_after) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CCDSpectrometerClass::pipe_factory_after
}
//--------------------------------------------------------
/**
 *	Method     : CCDSpectrometerClass::command_factory()
 *	Description: Create the command object(s)
 *                and store them in the command list
 */
//--------------------------------------------------------
void CCDSpectrometerClass::command_factory()
{
	/*----- PROTECTED REGION ID(CCDSpectrometerClass::command_factory_before) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CCDSpectrometerClass::command_factory_before


	//	Command Start
	StartClass	*pStartCmd =
		new StartClass("Start",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pStartCmd);

	//	Command Stop
	StopClass	*pStopCmd =
		new StopClass("Stop",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pStopCmd);

	//	Command Reset
	ResetClass	*pResetCmd =
		new ResetClass("Reset",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pResetCmd);

	//	Command StartSingle
	StartSingleClass	*pStartSingleCmd =
		new StartSingleClass("StartSingle",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pStartSingleCmd);

	//	Command StartContinuous
	StartContinuousClass	*pStartContinuousCmd =
		new StartContinuousClass("StartContinuous",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pStartContinuousCmd);

	//	Command GetLastData
	GetLastDataClass	*pGetLastDataCmd =
		new GetLastDataClass("GetLastData",
			Tango::DEV_LONG, Tango::DEVVAR_DOUBLEARRAY,
			"number of frames to return",
			"",
			Tango::OPERATOR);
	command_list.push_back(pGetLastDataCmd);

	//	Command AcquireNoise
	AcquireNoiseClass	*pAcquireNoiseCmd =
		new AcquireNoiseClass("AcquireNoise",
			Tango::DEV_LONG, Tango::DEV_VOID,
			"Number of frames to acquire as noise",
			"",
			Tango::OPERATOR);
	command_list.push_back(pAcquireNoiseCmd);

	//	Command SaveReferenceData
	SaveReferenceDataClass	*pSaveReferenceDataCmd =
		new SaveReferenceDataClass("SaveReferenceData",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pSaveReferenceDataCmd);

	//	Command GetData
	GetDataClass	*pGetDataCmd =
		new GetDataClass("GetData",
			Tango::DEVVAR_LONGARRAY, Tango::DEVVAR_DOUBLEARRAY,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pGetDataCmd);

	//	Command GetGaussFitCenter
	GetGaussFitCenterClass	*pGetGaussFitCenterCmd =
		new GetGaussFitCenterClass("GetGaussFitCenter",
			Tango::DEVVAR_LONGARRAY, Tango::DEVVAR_DOUBLEARRAY,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pGetGaussFitCenterCmd);

	//	Command GetGaussFitWidth
	GetGaussFitWidthClass	*pGetGaussFitWidthCmd =
		new GetGaussFitWidthClass("GetGaussFitWidth",
			Tango::DEVVAR_LONGARRAY, Tango::DEVVAR_DOUBLEARRAY,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pGetGaussFitWidthCmd);

	//	Command GetFWHMCenter
	GetFWHMCenterClass	*pGetFWHMCenterCmd =
		new GetFWHMCenterClass("GetFWHMCenter",
			Tango::DEVVAR_LONGARRAY, Tango::DEVVAR_DOUBLEARRAY,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pGetFWHMCenterCmd);

	//	Command GetFWHMWidth
	GetFWHMWidthClass	*pGetFWHMWidthCmd =
		new GetFWHMWidthClass("GetFWHMWidth",
			Tango::DEVVAR_LONGARRAY, Tango::DEVVAR_DOUBLEARRAY,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pGetFWHMWidthCmd);

	//	Command ExportWavelength
	ExportWavelengthClass	*pExportWavelengthCmd =
		new ExportWavelengthClass("ExportWavelength",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pExportWavelengthCmd);

	//	Command GetArea
	GetAreaClass	*pGetAreaCmd =
		new GetAreaClass("GetArea",
			Tango::DEVVAR_LONGARRAY, Tango::DEVVAR_DOUBLEARRAY,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pGetAreaCmd);

	//	Command GetWave
	GetWaveClass	*pGetWaveCmd =
		new GetWaveClass("GetWave",
			Tango::DEVVAR_LONGARRAY, Tango::DEVVAR_DOUBLEARRAY,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pGetWaveCmd);

	//	Command GetLambda
	GetLambdaClass	*pGetLambdaCmd =
		new GetLambdaClass("GetLambda",
			Tango::DEVVAR_LONGARRAY, Tango::DEVVAR_DOUBLEARRAY,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pGetLambdaCmd);

	/*----- PROTECTED REGION ID(CCDSpectrometerClass::command_factory_after) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CCDSpectrometerClass::command_factory_after
}

//===================================================================
//	Dynamic attributes related methods
//===================================================================

//--------------------------------------------------------
/**
 * method : 		CCDSpectrometerClass::create_static_attribute_list
 * description : 	Create the a list of static attributes
 *
 * @param	att_list	the created attribute list
 */
//--------------------------------------------------------
void CCDSpectrometerClass::create_static_attribute_list(std::vector<Tango::Attr *> &att_list)
{
	for (unsigned long i=0 ; i<att_list.size() ; i++)
	{
		std::string att_name(att_list[i]->get_name());
		transform(att_name.begin(), att_name.end(), att_name.begin(), ::tolower);
		defaultAttList.push_back(att_name);
	}

	TANGO_LOG_INFO << defaultAttList.size() << " attributes in default list" << std::endl;

	/*----- PROTECTED REGION ID(CCDSpectrometerClass::create_static_att_list) ENABLED START -----*/
	/* clang-format on */
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CCDSpectrometerClass::create_static_att_list
}


//--------------------------------------------------------
/**
 * method : 		CCDSpectrometerClass::erase_dynamic_attributes
 * description : 	delete the dynamic attributes if any.
 *
 * @param	devlist_ptr	the device list pointer
 * @param	list of all attributes
 */
//--------------------------------------------------------
void CCDSpectrometerClass::erase_dynamic_attributes(const Tango::DevVarStringArray *devlist_ptr, std::vector<Tango::Attr *> &att_list)
{
	Tango::Util *tg = Tango::Util::instance();

	for (unsigned long i=0 ; i<devlist_ptr->length() ; i++)
	{
		Tango::DeviceImpl *dev_impl = tg->get_device_by_name(((std::string)(*devlist_ptr)[i]).c_str());
		CCDSpectrometer *dev = static_cast<CCDSpectrometer *> (dev_impl);

		std::vector<Tango::Attribute *> &dev_att_list = dev->get_device_attr()->get_attribute_list();
		std::vector<Tango::Attribute *>::iterator ite_att;
		for (ite_att=dev_att_list.begin() ; ite_att != dev_att_list.end() ; ++ite_att)
		{
			std::string att_name((*ite_att)->get_name_lower());
			if ((att_name == "state") || (att_name == "status"))
				continue;
			std::vector<std::string>::iterator ite_str = find(defaultAttList.begin(), defaultAttList.end(), att_name);
			if (ite_str == defaultAttList.end())
			{
				TANGO_LOG_INFO << att_name << " is a UNWANTED dynamic attribute for device " << (*devlist_ptr)[i] << std::endl;
				Tango::Attribute &att = dev->get_device_attr()->get_attr_by_name(att_name.c_str());
				dev->remove_attribute(att_list[att.get_attr_idx()], true, false);
				--ite_att;
			}
		}
	}
	/*----- PROTECTED REGION ID(CCDSpectrometerClass::erase_dynamic_attributes) ENABLED START -----*/
	/* clang-format on */
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	CCDSpectrometerClass::erase_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Method     : CCDSpectrometerClass::get_attr_object_by_name()
 *	Description: returns Tango::Attr * object found by name
 */
//--------------------------------------------------------
Tango::Attr *CCDSpectrometerClass::get_attr_object_by_name(std::vector<Tango::Attr *> &att_list, std::string attname)
{
	std::vector<Tango::Attr *>::iterator it;
	for (it=att_list.begin() ; it<att_list.end() ; ++it)
		if ((*it)->get_name()==attname)
			return (*it);
	//	Attr does not exist
	return NULL;
}


/*----- PROTECTED REGION ID(CCDSpectrometerClass::Additional Methods) ENABLED START -----*/
/* clang-format on */
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	CCDSpectrometerClass::Additional Methods
} //	namespace
