#ifndef __TBUFFER_H
#define __TBUFFER_H

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <math.h>	//for NAN(18.04)
//#include <bits/nan.h>	//for NAN(14.04)
#include <stdint.h>

#define BUF_RT_NO_FILL		0
#define BUF_RT_FILL_1		1
#define BUF_RT_FILL_PREV	2
#define BUF_RT_FILL_0		3

#define FILL_ZERO	0x00000000
#define FILL_NAN	NAN
#define FILL_ONE_CHAR		(char)0xf
#define FILL_ONE_SHORT		(short)0xfff
#define FILL_ONE_LONG		(int32_t)0xfffffff
#define FILL_ONE_LONG_LONG	(int64_t)(0xfffffffffffffffULL)

#define BUF_GET_LAST_MODE		0
#define BUF_GET_BUNCH_MODE		1
#define BUF_GET_TIME_MODE		2
#define BUF_GET_LAST_UNIQUE_MODE	3
#define BUF_GET_BUNCH_UNIQUE_MODE	4
#define BUF_GET_TIME_UNIQUE_MODE	5
#define BUF_GET_LAST_BUNCH_MODE	6
#define BUF_GET_BUNCH_BUNCH_MODE	7
#define BUF_GET_TIME_BUNCH_MODE	8

#define	BUF_FILT_STD			0x00000001
#define	BUF_FILT_MEDIAN			0x00000002
#define	BUF_FILT_KALMAN			0x00000004

#define MAX_MEDIAN_SAMPLE		4096

//#define DEBUG_BUFF	1

#ifdef DEBUG_BUFF
#define PRINT(...) (DEBUG_BUFF ? (printf("%s: ", __func__), \
printf(__VA_ARGS__)) : 0)
#else
#define PRINT(...)
#endif

namespace tbuffer
{

typedef enum
{
	BUF_RT_LINEAR,
	BUF_RT_CIRCULAR
} bufRtMode_t;

typedef enum
{ 
	typeChar, 
	typeShort, 
	typeLong,
	typeFloat, 
	typeDouble,
	typeLongLong,
	typeCharArray,
	typeShortArray,
	typeLongArray,
	typeFloatArray,
	typeDoubleArray,
	typeLongLongArray
} type_t;

/*template <typename T>
class _buffer_data
{
	uint32_t bunchnr;	// bunch number -unambiguous-
	T data;			// data
	struct timeval time;	// timestamp
};*/

/*template <typename T>
class _buffer_data
{
	T data;			// data
};*/

/*class _bunch_time
{
	uint32_t bunchnr;	// bunch number -unambiguous-
	struct timeval time;	// timestamp
};*/

template <typename T>
class _data_array
{
	T *data;		// data
	size_t count;
public:
	_data_array(size_t num){data=new T[num]; count=num;}
	_data_array(T *val, size_t num)
	{
		data=new T[num];
		count=num;
		memcpy(data, val, sizeof(T)*num);
//		cout<<"copyed num="<<num<<" in _data_array, data[0]="<<data[0]<<" data[1]="<<data[1]<<" data[2]="<<data[2]<<" data[1000]="<<data[1000]<<" data[1001]="<<data[1001]<<" data[1002]="<<data[1002]<<endl;
	}
	_data_array(){data=NULL; count=0;}
	~_data_array(){if(data) delete [] data;}
	T& operator[] (const size_t& i){ return data[i]; }
	//T& operator new [] (const int& i){ return data[i]; }
	//const _data_array& operator= (const _data_array<T>&){ _data_array<T> ret(count); memcpy(ret.data, data, sizeof(T)*count); return ret; }
	const _data_array& operator= (const _data_array<T>& arg)
	{
		if(count != arg.count)
		{
			delete [] data;
			data=new T[arg.count];
			count=arg.count;
		}
		memcpy(data, arg.data, sizeof(T)*arg.count); return *this;
	}
	const _data_array& operator= (const T& arg)
	{
		if(data)
			delete [] data;
		data=new T[1];
		count=1;
		memcpy(data, &arg, sizeof(T)*1);
		return *this;
	}
	size_t extract_data(T* arg)
	{
		//cout << __func__<<": entering, count="<<count<<endl;
		if(count > 0)
			memcpy(arg, &data[0], sizeof(T)*count);
		//cout << __func__<<": after memcpy, count="<<count<<" data[0]="<<data[0]<<" data[1]="<<data[1]<<" data[2]="<<data[2]<<" data[1000]="<<data[1000]<<" data[1001]="<<data[1001]<<" data[1002]="<<data[1002]<<endl;
		return count;
	}
	size_t get_count()
	{
		return count;
	}
};


template<typename T>
struct identity { typedef T type; };

//-----------------------------------------------------------------------------------------------------------------

int tv_compare(struct timeval ts1, struct timeval ts2);

template <typename T>
class buffer_t
{
private:
	long  size;	// size of circular buffer
	type_t type;	//type of data
	long idx;	// index inside the buffer
	unsigned long counter;	// number of push executed
	unsigned long mode;	// flag that enables circular buffer, when 0 and the buffer is
			// full, push will return false
	unsigned long fill_hole;// is a bunch hole happens, this flag determines behaviour:
			// 0 = no fill
			// 1 = fill with 0xffffffff or NaN
			// 2 = fill with previous value inside the buffer
			// 3 = fill with 0
	//T *buffer_data;		// data
	uint32_t bunchnr[100000];	// bunch number -unambiguous-
	T data[100000];			// data
	timeval time[100000];	// timestamp
	//()()()()()()()()()()()()()()() KALMAN ()()()()()()()()()()()()()()()
	double sigmav;		//observation noise covariance
	double sigmaw;		//observation noise covariance
	double P;		//error covariance
	double x;		//real state
	double x_;		//predicted state
	double z;		//observation

type_t get_type(T foo)
{
	return m_get_type(foo);
}

public:
//template <typename TL>
type_t get_type()
{
	return get_type(identity<T>());
}
void fill0(T *val)
{
	*val = 0;
}
void fill1(T *val)
{
	fill1(val, identity<T>());
}
private:
template<typename TL>
type_t get_type(identity<TL>)
{
	return -1;
}

type_t get_type(identity<int8_t>)
{
	return typeChar;
}
type_t get_type(identity<int16_t>)
{
	return typeShort;
}
type_t get_type(identity<int32_t>)
{
	return typeLong;
}
type_t get_type(identity<int64_t>)
{
	return typeLongLong;
}
type_t get_type(identity<float>)
{
	return typeFloat;
}
type_t get_type(identity<double>)
{
	return typeDouble;
}
type_t get_type(identity<_data_array<int8_t> >)
{
	return typeCharArray;
}
type_t get_type(identity<_data_array<int16_t> >)
{
	return typeShortArray;
}
type_t get_type(identity<_data_array<int32_t> >)
{
	return typeLongArray;
}
type_t get_type(identity<_data_array<int64_t> >)
{
	return typeLongLongArray;
}
type_t get_type(identity<_data_array<float> >)
{
	return typeFloatArray;
}
type_t get_type(identity<_data_array<double> >)
{
	return typeDoubleArray;
}

template<typename TL>
void fill1(TL *val, identity<TL>)
{
	//TODO: data_array!!!!!!!!!!!!!!!!!!!!!!
}
void fill1(int8_t *val, identity<int8_t>)
{
	*val = FILL_ONE_CHAR;
}
void fill1(int16_t *val, identity<int16_t>)
{
	*val = FILL_ONE_SHORT;
}
void fill1(int32_t *val, identity<int32_t>)
{
	*val = FILL_ONE_LONG;
}
void fill1(int64_t *val, identity<int64_t>)
{
	*val = FILL_ONE_LONG_LONG;
}
void fill1(float *val, identity<float>)
{
	*val = FILL_NAN;
}
void fill1(double *val, identity<double>)
{
	*val = FILL_NAN;
}
void fill1(_data_array<int8_t> *val, identity<_data_array<int8_t> >)
{
	*val = _data_array<int8_t>(FILL_ONE_CHAR);
}
void fill1(_data_array<int16_t> *val, identity<_data_array<int16_t> >)
{
	*val = _data_array<int16_t>(FILL_ONE_SHORT);
}
void fill1(_data_array<int32_t> *val, identity<_data_array<int32_t> >)
{
	*val = _data_array<int32_t>(FILL_ONE_LONG);
}
void fill1(_data_array<int64_t> *val, identity<_data_array<int64_t> >)
{
	*val = _data_array<int64_t>(FILL_ONE_LONG_LONG);
}
void fill1(_data_array<float> *val, identity<_data_array<float> >)
{
	*val = _data_array<float>(FILL_NAN);
}
void fill1(_data_array<double> *val, identity<_data_array<double> >)
{
	*val = _data_array<double>(FILL_NAN);
}


public:
/***********************************
 *	allocates memory
 ***********************************/
int init_rt_buffer(long numsamples)
{

	//unsigned long type_size;
	//PRINT("entering... ver: %s %s file: %s\n", version_string, release_string, __FILE__rev);
	//trick to have release string in the binary, but I don't want the printf to be executed

	type = get_type();	//TODO
	//PRINT("setting type=%d\n",buffer.type);
	if(numsamples == 0)
		return -1;
	
	// allocates memory
	/*buffer_data = new T[numsamples];
	if (!buffer_data) 
		return -ENOMEM;*/
#ifdef _DYNAMIC_ALLOC
	bunchnr = NULL;
	data = NULL;
	time = NULL;

	bunchnr = new uint32_t[numsamples];	// bunch number -unambiguous-
	data = new T[numsamples];			// data
	time = new timeval[numsamples];	// timestamp
	if (!bunchnr || !data || !time) 
		return -ENOMEM;
// clean buffer
	//memset(buffer_data, 0, sizeof(buffer_data)*numsamples);
	memset(bunchnr, 0, sizeof(uint32_t)*numsamples);
	memset(data, 0, sizeof(T)*numsamples);
	memset(time, 0, sizeof(timeval)*numsamples);
#else
	memset(bunchnr, 0, sizeof(uint32_t)*100000);
	memset(data, 0, sizeof(T)*100000);
	memset(time, 0, sizeof(timeval)*100000);
#endif
// init size
	size = numsamples;
	
// init index
	idx = 0;
	counter = 0;
	mode = BUF_RT_CIRCULAR;
	fill_hole = BUF_RT_NO_FILL;
	
	return 0;
}
	
/***********************************
 *	 free memory
 ***********************************/
int delete_rt_buffer()
{
// frees memory
	//delete [] buffer_data;
#ifdef _DYNAMIC_ALLOC
	delete [] bunchnr;
	delete [] data;
	delete [] time;
#endif
	return 0;
}

/***********************************
 *	 reset buffer
 ***********************************/
int reset_rt_buffer()
{
// init index
	idx = 0;
	counter = 0;
	return 0;
}
	
/***********************************
 *	push data into buffer
 ***********************************/
int push_data(T *value, timeval timestamp, uint32_t bunchnumber)
{	
	struct timeval now;
	struct timezone tz;

//manage -full- condition
	if((counter < size)||(mode == BUF_RT_CIRCULAR))
	{

//buffer not full or circular => hole check
		if(counter!=0)	//first push skip hole check
		{
			long prev_idx = (idx-1+size)%size;
//found a hole
			while(((bunchnumber > bunchnr[prev_idx]) && (bunchnumber!=(bunchnr[prev_idx])+1)&&((counter != size)||(mode == BUF_RT_CIRCULAR)))&&(fill_hole!=BUF_RT_NO_FILL)){
				PRINT("data pushed pos: %li\n", idx);
				switch(fill_hole)
				{
					case BUF_RT_FILL_1: //fill with 0xffffffff, NAN
						fill1(&data[idx]);
						gettimeofday(&now,&tz);
						(time[idx]).tv_sec = now.tv_sec;
						(time[idx]).tv_usec = now.tv_usec;
						/*printf("pushed: %d, %d, %d-%d\n", FILL_ONE_LONG, (bunchnr_ptr[prev_idx])+1, now.tv_sec, now.tv_usec);*/
						PRINT("fill policy: added 1 on pos=%li\n", idx);
					break;
					case BUF_RT_FILL_0: //fill with 0
						fill0(&data[idx]);
						gettimeofday(&now,&tz);
						(time[idx]).tv_sec = now.tv_sec;
						(time[idx]).tv_usec = now.tv_usec;
						/*printf("pushed: %d, %d, %d-%d\n", FILL_ONE_LONG, (bunchnr_ptr[prev_idx])+1, now.tv_sec, now.tv_usec);*/
						PRINT("fill policy: added 0 on pos=%li\n", idx);
					break;
					case BUF_RT_FILL_PREV: //fill with previous value
						data[idx] = data[prev_idx];
						(time[idx]).tv_sec = (time[prev_idx]).tv_sec;
						(time[idx]).tv_usec = (time[prev_idx]).tv_usec;
						/*PRINT("pushed: %f, %d, %d-%d\n", data_ptr[prev_idx], (bunchnr_ptr[prev_idx])+1, time_ptr[prev_idx].tv_sec, time_ptr[prev_idx].tv_usec);*/
						PRINT("fill policy: added previous  on pos=%li\n", idx);
					break;
				}
				bunchnr[idx] = bunchnr[prev_idx]+1;
				//TODO: why counter limited to size?
				//this->counter = (this->counter==this->size)?this->size:this->counter+1;
				counter++;
				idx = ((idx + 1) % size);
				prev_idx = (idx-1+size)%size;
			}
		}
		else
		{
			//PRINT("no holes check\n");
		}
//insertion
		if((counter != size)||(mode == BUF_RT_CIRCULAR))
		{
			data[idx] = *value;
			(time[idx]).tv_sec = timestamp.tv_sec;
			(time[idx]).tv_usec = timestamp.tv_usec;
			bunchnr[idx] = bunchnumber;
			//TODO: why counter limited to size?
			//this->counter = (this->counter==this->size)?this->size:this->counter+1;
			counter++;
			//TODO: memory barrier here?
#ifndef __KERNEL__
#if __BYTE_ORDER == __BIG_ENDIAN	//powerpc
			asm volatile("lwsync");
#else
			asm volatile("mfence");
#endif
#endif
			idx = ((idx + 1) % size);
			return 0;
		}
		else{			
			PRINT("no push after holes check: buffer is full counter=%li size=%li\n",counter,size);
			return -1;
		}
	}
	else{			
		PRINT("no push: buffer is full counter=%li size=%li\n",counter,size);
		return -1;
	}
};

/***********************************
 *	extract data from buffer
 ***********************************/
int pop_data(unsigned char unique, long idx[2], T *data_buf)
{
	long i = idx[0];
	long h = idx[1];
	long size_tmp1 = 0;
	long size_tmp2 = 0;
	read_data(unique, idx, data_buf);
	if(i < h)
	{
		size_tmp1 = h-i;
	}
	else if(i > h)
	{
		size_tmp1 = this->size - i;		//i -> buffer size
		size_tmp2 = h;				//0 -> h
	}
	idx = ((this->idx)-(size_tmp1 + size_tmp2)+size)%size;
	return 0;
};

/*int read_data(unsigned char unique, long idx[2], T *data_buf)
{
	return read_data(unsigned char unique, long idx[2], T *data_buf);
};
private:
template<typename TL>
long get_data_size(identity<TL>)
{
	return this->size;
}
template<typename TL>
long get_data_size(identity<_data_array<TL> >)
{
	long ret_size=0;
	for(int i=0; i<size; i++)
	{
		ret_size += data[i].count;
	}
	return ret_size;
}

private:*/
/***********************************
 *	read data from buffer
 ***********************************/
int read_data(unsigned char unique, long idx[2], T *data_buf){

	long i = idx[0];
	long h = idx[1];
	long size_tmp1 = 0;
	long size_tmp2 = 0;
	int k = 0, j=0;
	long ns;
	unsigned long last_bunchnr = -1;
	T *data_ptr = &data[0];
	uint32_t *bunchnr_ptr = (uint32_t *)&bunchnr[0];
	if(i < 0 || h < 0)
		return -1;
	//long ns = (h-i+this->size)%this->size;
	PRINT("from idx[0] = %li, to idx[1] = %li\n", idx[0], idx[1]);
	/*while(k<ns){	//TODO: memcpy
		switch(this->type)
		{
			case typeChar:
				((char*)data_buf)[k] = ((char*)data_ptr)[i];
				break;
			case typeShort:
				((short*)data_buf)[k] = ((short*)data_ptr)[i];
				break;
			case typeLong:
				((long*)data_buf)[k] = ((long*)data_ptr)[i];
				break;
			case typeFloat:
				((float*)data_buf)[k] = ((float*)data_ptr)[i];
				break;
			case typeDouble:
				((double*)data_buf)[k] = ((double*)data_ptr)[i];
				break;
			case typeLongLong:
				((long long*)data_buf)[k] = ((long long*)data_ptr)[i];
				break;
			default:
				break;
		}
		i+=1;
		if(i==this->size)i=0;
		k+=1;
		PRINT("saving index %d to pos %d\n", i, k);
	}*/
	if(unique!=1)
	{
		//case 1: just memcpy between indexes
		if(i <= h)
		{
			size_tmp1 = h-i+1;
			switch(type)
			{
				case typeChar:
				case typeShort:
					memcpy(&((T*)data_buf)[0], &((T*)data_ptr)[i], size_tmp1 * sizeof(T));
					break;
				case typeLong:
					memcpy(&((T*)data_buf)[0], &((T*)data_ptr)[i], size_tmp1 * sizeof(T));
					if(unique == 2)
						memcpy(&((T)data_buf)[size_tmp1], &bunchnr_ptr[i], size_tmp1 * sizeof(T));
					//PRINT("copied from %li, to %li\n", ((long*)data_ptr)[i],((long*)data_buf)[0]);
					break;
				case typeFloat:
				case typeDouble:
				case typeLongLong:
					memcpy(&((T*)data_buf)[0], &((T*)data_ptr)[i], size_tmp1 * sizeof(T));
					if(unique == 2)
					{
						for(k=0; k < size_tmp1; k++)
						{
							((T*)data_buf)[size_tmp1+k] =  (T)bunchnr_ptr[i+k];
						}
					}
					break;
				default:
					break;
			}
		}
		//case 2: memcpy from i -> buffer size and from 0 to h
		else if(i > h)
		{
			size_tmp1 = size - i;		//i -> buffer size
			size_tmp2 = h+1;				//0 -> h
			switch(type)
			{
				case typeChar:
				case typeShort:
					memcpy(&((T*)data_buf)[0], &((T*)data_ptr)[i], size_tmp1 * sizeof(T));
					memcpy(&((T*)data_buf)[size_tmp1], &((T*)data_ptr)[0], size_tmp2 * sizeof(T));
					break;
				case typeLong:
					memcpy(&((T*)data_buf)[0], &((T*)data_ptr)[i], size_tmp1 * sizeof(T));
					memcpy(&((T*)data_buf)[size_tmp1], &((T*)data_ptr)[0], size_tmp2 * sizeof(T));
					if(unique == 2)
					{
						memcpy(&((T*)data_buf)[size_tmp1+size_tmp2], &bunchnr_ptr[i], size_tmp1 * sizeof(T));
						memcpy(&((T*)data_buf)[size_tmp1+size_tmp2+size_tmp1], &bunchnr_ptr[0], size_tmp2 * sizeof(T));
					}
					break;
				case typeFloat:
				case typeDouble:
				case typeLongLong:
					memcpy(&((T*)data_buf)[0], &((T*)data_ptr)[i], size_tmp1 * sizeof(T));
					memcpy(&((T*)data_buf)[size_tmp1], &((T*)data_ptr)[0], size_tmp2 * sizeof(T));
					if(unique == 2)
					{
						for(k=0; k < size_tmp1; k++)
						{
							((T*)data_buf)[size_tmp1+size_tmp2+k] =  (T)bunchnr_ptr[i+k];
						}
						for(k=0; k < size_tmp2; k++)
						{
							((T*)data_buf)[size_tmp1+size_tmp2+size_tmp1+k] =  (T)bunchnr_ptr[k];
						}
					}
					break;
				default:
					break;
			}	
		}
	}
	else	//unique
	{
		//ns = (h-i+1+this->size)%this->size;
		if(h > i)
		{
			ns = h - i;
		}
		else if(h < i)
		{
			ns = this->size - i + h;
		}
		else
		{
			ns = 0;
		}
		ns+=1;
		PRINT("from idx[0] = %li, to idx[1] = %li, unique size=%li\n", idx[0], idx[1], ns);
		while(j<ns)
		{
			if(last_bunchnr == bunchnr_ptr[i])
			{
				i+=1;
				if(i==size)i=0;
				j++;
				PRINT("skipped bunchnr=%li, now i=%li, j=%i\n", last_bunchnr, i,j);
				continue;
			}
			else
			{
				last_bunchnr = bunchnr_ptr[i];
				j++;
				PRINT("now bunchnr=%li, i=%li,j=%i\n", last_bunchnr, i,j);
			}
			((T*)data_buf)[k] = ((T*)data_ptr)[i];
			i+=1;
			if(i==size)i=0;
			k+=1;
		}
	}
	return 0;
};

/***********************************
 *	read data from buffer
 ***********************************/
template <typename TL> int read_data(unsigned char unique, long idx[2], TL *data_buf, _data_array<TL> *foo){

	long i = idx[0];
	long h = idx[1];
	long size_tmp1 = 0;
	long size_tmp2 = 0;
	int k = 0, j=0;
	long ns;
	unsigned long last_bunchnr = -1;
	T *data_ptr = &data[0];
	uint32_t *bunchnr_ptr = (uint32_t *)&bunchnr[0];
	if(i < 0 || h < 0)
		return -1;
	//long ns = (h-i+this->size)%this->size;
	PRINT("from idx[0] = %li, to idx[1] = %li\n", idx[0], idx[1]);
	/*while(k<ns){	//TODO: memcpy
		switch(this->type)
		{
			case typeChar:
				((char*)data_buf)[k] = ((char*)data_ptr)[i];
				break;
			case typeShort:
				((short*)data_buf)[k] = ((short*)data_ptr)[i];
				break;
			case typeLong:
				((long*)data_buf)[k] = ((long*)data_ptr)[i];
				break;
			case typeFloat:
				((float*)data_buf)[k] = ((float*)data_ptr)[i];
				break;
			case typeDouble:
				((double*)data_buf)[k] = ((double*)data_ptr)[i];
				break;
			case typeLongLong:
				((long long*)data_buf)[k] = ((long long*)data_ptr)[i];
				break;
			default:
				break;
		}
		i+=1;
		if(i==this->size)i=0;
		k+=1;
		PRINT("saving index %d to pos %d\n", i, k);
	}*/
	if(unique!=1)
	{
		//case 1: just memcpy between indexes
		if(i <= h)
		{
			size_tmp1 = h-i+1;
			switch(type)
			{
				case typeCharArray:
				case typeShortArray:
				case typeLongArray:
				case typeFloatArray:
				case typeDoubleArray:
				case typeLongLongArray:
				{
					int z=0;
					//(unique == 2) not supported
					for(k=0; k < size_tmp1; k++)
					{
						PRINT("k=%d size_tmp1=%d\n", k, size_tmp1);
						int ret=data_ptr[i+k].extract_data(&data_buf[z]);
						z+=ret;
						PRINT("z=%d\n", z);
					}
					break;
				}
				default:
					break;
			}
		}
		//case 2: memcpy from i -> buffer size and from 0 to h
		else if(i > h)
		{
			size_tmp1 = size - i;		//i -> buffer size
			size_tmp2 = h+1;				//0 -> h
			switch(type)
			{
				case typeCharArray:
				case typeShortArray:
				case typeLongArray:
				case typeFloatArray:
				case typeDoubleArray:
				case typeLongLongArray:
				{
					int z=0;
					//(unique == 2) not supported
					for(k=0; k < size_tmp1; k++)
					{
						int ret=data_ptr[i+k].extract_data(&data_buf[z]);
						z+=ret;
					}
					for(k=0; k < size_tmp2; k++)
					{
						int ret=data_ptr[0+k].extract_data(&data_buf[z]);
						z+=ret;
					}
					break;
				}
				default:
					break;
			}
		}
	}
	else	//unique
	{
		//ns = (h-i+1+this->size)%this->size;
		if(h > i)
		{
			ns = h - i;
		}
		else if(h < i)
		{
			ns = this->size - i + h;
		}
		else
		{
			ns = 0;
		}
		ns+=1;
		PRINT("from idx[0] = %li, to idx[1] = %li, unique size=%li\n", idx[0], idx[1], ns);
		while(j<ns)
		{
			if(last_bunchnr == bunchnr_ptr[i])
			{
				i+=1;
				if(i==size)i=0;
				j++;
				PRINT("skipped bunchnr=%li, now i=%li, j=%i\n", last_bunchnr, i,j);
				continue;
			}
			else
			{
				last_bunchnr = bunchnr_ptr[i];
				j++;
				PRINT("now bunchnr=%li, i=%li,j=%i\n", last_bunchnr, i,j);
			}
			((T*)data_buf)[k] = ((T*)data_ptr)[i];
			i+=1;
			if(i==size)i=0;
			k+=1;
		}
	}
	return 0;
}

public:

#define ELEM_SWAP_L(a,b) { register int32_t t=(a);(a)=(b);(b)=t; }
#define ELEM_SWAP_D(a,b) { register double t=(a);(a)=(b);(b)=t; }

int read_data_filter(unsigned char unique, long idx[2], double x2, double x1, double x0, double *std, double *mean, void *median, double *kalman, int mask, void *skip)
{
	long i = idx[0];
	long h = idx[1];
	int ns;
	char to_be_skipped=0;

	uint32_t last_bunchnr = -1;
	int k = 0, j=0;
	int count=0;
	double val;
	double delta = 0.0;
	double M2 = 0.0;

	int low, high ;
	int median_idx;
	int middle, ll, hh;
	unsigned long type_size;
//	long *arr_l;
	int32_t arr_l[MAX_MEDIAN_SAMPLE];
//	double *arr_d;
	double arr_d[MAX_MEDIAN_SAMPLE];

	T *data_ptr = &data[0];
	uint32_t *bunchnr_ptr = (uint32_t *)&bunchnr[0];

	if(mean)
		*mean=0;
	if(std)
		*std=0;
	if(median)
	{
		switch(this->type)
		{
			case typeChar:
			case typeShort:
			case typeFloat:
			case typeLongLong:
				break;
			case typeLong:
			case typeDouble:
				*(T *)median=0;
				break;
		}
	}
	if(kalman)
		*kalman=0;

	if(mask == 0)	//if no mask selected
	{
		//return last (average of 1)
		mask = BUF_FILT_STD;
		i=h;
	}

	if(h > i)
		ns = h - i;
	else if(h < i)
		ns = size - i + h;
	else
		ns = 0;
	ns+=1;

	if(i < 0 || h < 0)
		return -1;
	//long ns = (h-i+this->size)%this->size;
	PRINT("from idx[0] = %li, to idx[1] = %li numsample=%i\n", idx[0], idx[1], ns);
	//std/mean algorithm due to Knuth
	if(mask & BUF_FILT_STD)
	{
		while(j<ns)
		{
			if(unique==1)
			{
				if(last_bunchnr == bunchnr_ptr[i])
				{
					i+=1;
					if(i==this->size)i=0;
					j++;
					continue;
				}
				else
				{
					last_bunchnr = bunchnr_ptr[i];
				}
			}

			switch(this->type)
			{
				case typeChar:
				case typeShort:
				case typeLong:
				case typeLongLong:
					val = x2 * ((T*)data_ptr)[i] * ((T*)data_ptr)[i] + x1 * ((T*)data_ptr)[i] + x0;
					to_be_skipped = (skip != NULL) ? ((((T*)data_ptr)[i] == *(T*)skip) ? 1 : 0) : 0;
					break;
				case typeFloat:
				case typeDouble:
					val = x2 * ((T*)data_ptr)[i] * ((T*)data_ptr)[i] + x1 * ((T*)data_ptr)[i] + x0;
					to_be_skipped = (skip != NULL) ? ((((T*)data_ptr)[i] == *(T*)skip) ? 1 : ((isnan(((T*)data_ptr)[i]) || isinf(((T*)data_ptr)[i])) ? 1 : 0)) : 0;
					break;
				case typeCharArray:
				case typeShortArray:
				case typeLongArray:
				case typeFloatArray:
				case typeDoubleArray:
				case typeLongLongArray:
					break;
				default:
					break;
			}

			i+=1;
			if(i==this->size)i=0;
			k+=1;
			j++;
			if(to_be_skipped == 1)
				continue;
			//std/mean algorithm due to Knuth
			count++;
			delta =  val - (*mean);
			*mean = *mean + delta/count;
			M2 = M2 + delta*(val - *mean);  // This expression uses the new value of mean
			PRINT("MEAN/STD: count=%d val=%f mean=%f\n", count, val, *mean);
		}
		if(count < 1)
		{
			return -1;
		}
		else if(count == 1)
		{
			*std = 0;
		}
		else
		{
			*std = M2/(count - 1);
			*std = sqrt(*std);
		}
	}
	if(mask & BUF_FILT_MEDIAN)
	{
		if(ns > MAX_MEDIAN_SAMPLE)
			return -1;
		switch(this->type)
		{
			case typeChar:
				type_size = sizeof(char);
				break;
			case typeShort:
				type_size = sizeof(short);
				break;
			case typeLong:
				type_size = sizeof(int32_t);
				break;
			case typeFloat:
				type_size = sizeof(float);
				break;
			case typeDouble:
				type_size = sizeof(double);
				break;
			case typeLongLong:
				type_size = sizeof(int64_t);
				break;
			default:
				type_size = 0;
				break;
		}



// allocates memory
		switch(this->type)
		{
			case typeChar:
			case typeShort:
			case typeFloat:
			case typeLongLong:
				break;
			case typeLong:
				/*arr_l = malloc(type_size*ns);
				if (!arr_l)
				{ 
					return -ENOMEM;
				}*/
				read_data(unique%2, idx, (void *)arr_l);	//not interested in bunchnumbers if unique=2 here
				break;
			case typeDouble:
				/*arr_d = malloc(type_size*ns);
				if (!arr_d)
				{ 
					return -ENOMEM;
				}*/
				read_data(unique%2, idx, (void *)arr_d);	//not interested in bunchnumbers if unique=2 here
				break;
		}

		switch(this->type)
		{
			case typeChar:
			case typeShort:
			case typeFloat:
			case typeLongLong:
				break;
			case typeLong:
				if(skip != NULL)
				{
					j=0;
					for(i=0; i<ns; i++)
					{
						if(arr_l[i] != *(int32_t*)skip)
						{
							arr_l[j++] = arr_l[i];
							PRINT("MEDIAN: skipping... arr_l[%d]=%li\n", j, arr_l[j-1]);
						}
						/*else	//to be skipped
							*/
						
					}
					ns = j;
				}
				low = 0 ; high = ns-1 ; median_idx = (low + high) / 2;
				for (;;) {
					PRINT("MEDIAN: low=%d high=%d median_idx=%d\n", low, high, median_idx);
					if (high <= low) /* One element only */
						goto found_median;
			
					if (high == low + 1) {  /* Two elements only */
						if (arr_l[low] > arr_l[high])
							ELEM_SWAP_L(arr_l[low], arr_l[high]);
							goto found_median;
					}
			
					/* Find median of low, middle and high items; swap into position low */
					middle = (low + high) / 2;
					if (arr_l[middle] > arr_l[high])    ELEM_SWAP_L(arr_l[middle], arr_l[high]) ;
					if (arr_l[low] > arr_l[high])       ELEM_SWAP_L(arr_l[low], arr_l[high]) ;
					if (arr_l[middle] > arr_l[low])     ELEM_SWAP_L(arr_l[middle], arr_l[low]) ;
			
					/* Swap low item (now in position middle) into position (low+1) */
					ELEM_SWAP_L(arr_l[middle], arr_l[low+1]) ;
			
					/* Nibble from each end towards middle, swapping items when stuck */
					ll = low + 1;
					hh = high;
					for (;;) {
						do ll++; while (arr_l[low] > arr_l[ll]) ;
						do hh--; while (arr_l[hh]  > arr_l[low]) ;
			
						if (hh < ll)
							break;
			
						ELEM_SWAP_L(arr_l[ll], arr_l[hh]) ;
					}
			
					/* Swap middle item (in position low) back into correct position */
					ELEM_SWAP_L(arr_l[low], arr_l[hh]) ;
			
					/* Re-set active partition */
					if (hh <= median_idx)
						low = ll;
					if (hh >= median_idx)
						high = hh - 1;
				}
				break;
			case typeDouble:
				if(skip != NULL)
				{
					j=0;
					for(i=0; i<ns; i++)
					{
						if(arr_d[i] != *(double*)skip)
						{
							arr_d[j++] = arr_d[i];
							PRINT("MEDIAN: skipping... arr_d[%d]=%f\n", j, arr_d[j-1]);
						}
						/*else	//to be skipped
							*/
						
					}
					ns = j;
				}
				low = 0 ; high = ns-1 ; median_idx = (low + high) / 2;
				for (;;) {
					PRINT("MEDIAN: low=%d high=%d median_idx=%d\n", low, high, median_idx);
					if (high <= low) /* One element only */
						goto found_median;
			
					if (high == low + 1) {  /* Two elements only */
						if (arr_d[low] > arr_d[high])
							ELEM_SWAP_D(arr_d[low], arr_d[high]);
							goto found_median;
					}
			
					/* Find median of low, middle and high items; swap into position low */
					middle = (low + high) / 2;
					if (arr_d[middle] > arr_d[high])    ELEM_SWAP_D(arr_d[middle], arr_d[high]) ;
					if (arr_d[low] > arr_d[high])       ELEM_SWAP_D(arr_d[low], arr_d[high]) ;
					if (arr_d[middle] > arr_d[low])     ELEM_SWAP_D(arr_d[middle], arr_d[low]) ;
			
					/* Swap low item (now in position middle) into position (low+1) */
					ELEM_SWAP_D(arr_d[middle], arr_d[low+1]) ;
			
					/* Nibble from each end towards middle, swapping items when stuck */
					ll = low + 1;
					hh = high;
					for (;;) {
						do ll++; while (arr_d[low] > arr_d[ll]) ;
						do hh--; while (arr_d[hh]  > arr_d[low]) ;
			
						if (hh < ll)
							break;
			
						ELEM_SWAP_D(arr_d[ll], arr_d[hh]) ;
					}
			
					/* Swap middle item (in position low) back into correct position */
					ELEM_SWAP_D(arr_d[low], arr_d[hh]) ;
			
					/* Re-set active partition */
					if (hh <= median_idx)
						low = ll;
					if (hh >= median_idx)
						high = hh - 1;
				}
				break;
		}
found_median:
		switch(this->type)
		{
			case typeChar:
			case typeShort:
			case typeFloat:
			case typeLongLong:
				break;
			case typeLong:
				*(double *)median = x2*arr_l[median_idx]*arr_l[median_idx] + x1*arr_l[median_idx] + x0;
				break;
			case typeDouble:
				*(double *)median = x2*arr_d[median_idx]*arr_d[median_idx] + x1*arr_d[median_idx] +x0;
				break;
		}
// frees memory
	switch(this->type)
	{
		case typeChar:
		case typeShort:
		case typeFloat:
		case typeLongLong:
			break;
		case typeLong:
		//	free(arr_l);
			break;
		case typeDouble:
		//	free(arr_d);
			break;
	}
	}
	return 0;
}


/****************************************
 *	read data calibrated from buffer
 *	each data is calibrated as:
 *	x2 * d^2 + x1 * d + x0
 *	return the number of data read
 ****************************************/
int read_data_calib(unsigned char unique, long idx[2], double *data_buf, double x2, double x1, double x0){

	long i = idx[0];
	long h = idx[1];
	int k = 0, j=0;
	int ns;
	uint32_t last_bunchnr = -1;
	 
	T *data_ptr = &data[0];
	uint32_t *bunchnr_ptr = (uint32_t *)&bunchnr[0];

	if(h > i)
		ns = h - i;
	else if(h < i)
		ns = this->size - i + h;
	else
		ns = 0;
	ns+=1;

	if(i < 0 || h < 0)
		return -1;
	//long ns = (h-i+this->size)%this->size;
	//PRINT("from idx[0] = %li, to idx[1] = %li numsample=%i\n", idx[0], idx[1], ns);
	while(j<ns)
	{
		if(unique == 1)
		{
			if(last_bunchnr == bunchnr_ptr[i])
			{
				i+=1;
				if(i==this->size)i=0;
				j++;
				continue;
			}
			else
			{
				last_bunchnr = bunchnr_ptr[i];
			}
		}
		switch(this->type)
		{
			case typeChar:
				data_buf[k] = x2 * ((int8_t*)data_ptr)[i] * ((int8_t*)data_ptr)[i] + x1 * ((int8_t*)data_ptr)[i] + x0;
				break;
			case typeShort:
				data_buf[k] = x2 * ((int16_t*)data_ptr)[i] * ((int16_t*)data_ptr)[i] + x1 * ((int16_t*)data_ptr)[i] + x0;
				break;
			case typeLong:
				data_buf[k] = x2 * ((int32_t*)data_ptr)[i] * ((int32_t*)data_ptr)[i] + x1 * ((int32_t*)data_ptr)[i] + x0;
				break;
			case typeFloat:
				data_buf[k] = x2 * ((float*)data_ptr)[i] * ((float*)data_ptr)[i] + x1 * ((float*)data_ptr)[i] + x0;
				break;
			case typeDouble:
				data_buf[k] = x2 * ((double*)data_ptr)[i] * ((double*)data_ptr)[i] + x1 * ((double*)data_ptr)[i] + x0;
				break;
			case typeLongLong:
				data_buf[k] = x2 * ((int64_t*)data_ptr)[i] * ((int64_t*)data_ptr)[i] + x1 * ((int64_t*)data_ptr)[i] + x0;
				break;
			case typeCharArray:
			case typeShortArray:
			case typeLongArray:
			case typeFloatArray:
			case typeDoubleArray:
			case typeLongLongArray:
			/*{
				int ret=data_ptr[i].extract_data(&data_buf[k]);
				for(int zz=i; zz<i+ret; zz++)
				{
					data_buf[zz] = x2 * (data_buf)[zz] * (data_buf)[zz] + x1 * (data_buf)[zz] + x0;
				}
				break;
			}*/
			default:
				break;
		}
		if(unique == 2)
			data_buf[k+ns] = (double)bunchnr_ptr[i];	//concatenate bunchnumbers
		i+=1;
		if(i==this->size)i=0;
		k+=1;
		j++;
		//PRINT("saving index %ld to pos %d\n", i, k);
	}
	return 0;
};
//#endif
//#ifndef __KERNEL__	//TODO: __floatdidf
//std/mean algorithm due to Knuth
template <typename TL> int read_std_calib(unsigned char unique, long idx[2], double x2, double x1, double x0, double *std, double *mean)
{
	long i = idx[0];
	long h = idx[1];
	int k = 0, j=0;
	int ns;
	uint32_t last_bunchnr = -1;

	int count=0;
	double val;
	double delta = 0;
	double M2 = 0;

	TL *data_ptr = &data[0];
	uint32_t *bunchnr_ptr = (uint32_t *)&bunchnr[0];

	*mean=0;
	*std=0;

	if(h > i)
		ns = h - i;
	else if(h < i)
		ns = this->size - i + h;
	else
		ns = 0;
	ns+=1;

	if(i < 0 || h < 0)
		return -1;
	//long ns = (h-i+this->size)%this->size;
	PRINT("from idx[0] = %li, to idx[1] = %li numsample=%i\n", idx[0], idx[1], ns);
	while(j<ns)
	{
		if(unique==1)
		{
			if(last_bunchnr == bunchnr_ptr[i])
			{
				i+=1;
				if(i==this->size)i=0;
				j++;
				continue;
			}
			else
			{
				last_bunchnr = bunchnr_ptr[i];
			}
		}
		switch(this->type)
		{
			case typeChar:
				val = x2 * ((int8_t*)data_ptr)[i] * ((int8_t*)data_ptr)[i] + x1 * ((int8_t*)data_ptr)[i] + x0;
				break;
			case typeShort:
				val = x2 * ((int16_t*)data_ptr)[i] * ((int16_t*)data_ptr)[i] + x1 * ((int16_t*)data_ptr)[i] + x0;
				break;
			case typeLong:
				val = x2 * ((int32_t*)data_ptr)[i] * ((int32_t*)data_ptr)[i] + x1 * ((int32_t*)data_ptr)[i] + x0;
				break;
			case typeFloat:
				val = x2 * ((float*)data_ptr)[i] * ((float*)data_ptr)[i] + x1 * ((float*)data_ptr)[i] + x0;
				break;
			case typeDouble:
				val = x2 * ((double*)data_ptr)[i] * ((double*)data_ptr)[i] + x1 * ((double*)data_ptr)[i] + x0;
				break;
			case typeLongLong:
				val = x2 * ((int64_t*)data_ptr)[i] * ((int64_t*)data_ptr)[i] + x1 * ((int64_t*)data_ptr)[i] + x0;
				break;
			case typeCharArray:
			case typeShortArray:
			case typeLongArray:
			case typeFloatArray:
			case typeDoubleArray:
			case typeLongLongArray:
			{
				/*int ret=data_ptr[i].extract_data(&data_buf[k]);
				for(int zz=i; zz<i+ret; zz++)
				{
					data_ptr[zz] = x2 * ((T*)data_ptr)[zz] * ((T*)data_ptr)[zz] + x1 * ((T*)data_ptr)[zz] + x0;
				}*/
				break;
			}
			default:
				break;
		}
		i+=1;
		if(i==this->size)i=0;
		k+=1;
		j++;

		//std/mean algorithm due to Knuth
		count++;
		delta =  val - *mean;
		*mean = *mean + delta/count;
		M2 = M2 + delta*(val - *mean);  // This expression uses the new value of mean
	}
	if(count < 1)
	{
		return -1;
	}
	else if(count == 1)
	{
		*std = 0;
	}
	else
	{
		*std = M2/(count - 1);
		*std = sqrt(*std);
	}
	return 0;
}

int init_kalman(double sigmav, double sigmaw, double P)
{
	long idx;
	T *data_ptr = &data[0];

	this->sigmav = sigmav;
	this->sigmaw = sigmaw;
	this->P = P;			//initial error covariance

	//initial predicted state
	if(this->counter != 0)
	{
		idx = (this->idx-1+this->size)%this->size;
		switch(this->type)
		{
			case typeChar:
				this->x_ = ((char*)data_ptr)[idx];
				break;
			case typeShort:
				this->x_ = ((short*)data_ptr)[idx];
				break;
			case typeLong:
				this->x_ = ((int32_t*)data_ptr)[idx];
				break;
			case typeFloat:
				this->x_ = ((float*)data_ptr)[idx];
				break;
			case typeDouble:
				this->x_ = ((double*)data_ptr)[idx];
				break;
			case typeLongLong:
				this->x_ = ((int64_t*)data_ptr)[idx];
				break;
			default:
				break;
		}
	}
	else
	{
		this->x_ = 0;
	}
	return 0;
}

int read_kalman(double *data_buf, unsigned int num_samples)
{
	unsigned long num;
	long idx[2];
	unsigned int i;
	long ind;
	T *data_ptr = this->data;
	double Q,R;
	double S;
	double K;

	Q = this->sigmav * this->sigmav;
	R = this->sigmaw * this->sigmaw;
	num = get_last_idx(num_samples, 0, idx);
	PRINT("read_kalman entering: Q=%f R=%f P=%.11e num_samples=%ld x_=%f\n", Q, R, this->P, num, this->x_);
	if(num > 0)
	{
		ind = idx[0];
		for(i=0; i<num; i++)
		{
			this->P = this->P + Q;	//Prediction of the plant covariance
			S = this->P + R;	//Innovation Covariance
			K = this->P/S;	//Kalman's gain

			switch(this->type)
			{
				case typeChar:
					this->z = ((char*)data_ptr)[ind];
					break;
				case typeShort:
					this->z = ((short*)data_ptr)[ind];
					break;
				case typeLong:
					this->z = ((int32_t*)data_ptr)[ind];
					break;
				case typeFloat:
					this->z = ((float*)data_ptr)[ind];
					break;
				case typeDouble:
					this->z = ((double*)data_ptr)[ind];
					break;
				case typeLongLong:
					this->z = ((int64_t*)data_ptr)[ind];
					break;
				default:
					break;
			}
			this->x_ = this->x_ + K * (this->z - this->x_);
			this->P = (1 - K) * this->P;
			ind = ((ind + 1) % this->size);
			PRINT("read_kalman filtering: P=%.13e K=%.11e i=%d/%ld index=%ld | z=%f  x_=%.20f\n", this->P,K, i, num, ind, this->z, this->x_);
		}
	}
	else
		return -1;

	*data_buf = this->x_;
	return 0;
}

/***************************************
 *	read last data, last timestamp *
 *	and last bunch number          *
 *	return -1 on error, 0 if OK    *
 ***************************************/
int read_last(void *dat, struct timeval *timestamp, uint32_t *bunchnumber)
{
	long idx;
	T *data_ptr = &data[0];
	struct timeval *time_ptr = (struct timeval *)&time[0];
	uint32_t *bunchnr_ptr = (uint32_t *)&bunchnr[0];
	if(this->counter != 0)
	{
		PRINT("read_last this->counter=%ld\n",this->counter);
		idx = (this->idx-1+this->size)%this->size;
		switch(this->type)
		{
			case typeChar:
				memcpy(&((char*)dat)[0], &((char*)data_ptr)[idx], 1 * sizeof(char));
				break;
			case typeShort:
				memcpy(&((short*)dat)[0], &((short*)data_ptr)[idx], 1 * sizeof(short));
				break;
			case typeLong:
				memcpy(&((int32_t*)dat)[0], &((int32_t*)data_ptr)[idx], 1 * sizeof(int32_t));
				break;
			case typeFloat:
				memcpy(&((float*)dat)[0], &((float*)data_ptr)[idx], 1 * sizeof(float));
				break;
			case typeDouble:
				memcpy(&((double*)dat)[0], &((double*)data_ptr)[idx], 1 * sizeof(double));
				//PRINT("read_last: copying %f in %f\n", ((double*)data_ptr)[idx], *(double*)dat);
				break;
			case typeLongLong:
				memcpy(&((int64_t*)dat)[0], &((int64_t*)data_ptr)[idx], 1 * sizeof(int64_t));
				break;
			default:
				break;
		}
		*bunchnumber = bunchnr_ptr[idx];
		*timestamp = time_ptr[idx];
	}
	else
	{
		return -1;
	}
	return 0;
}

	
/***********************************
 *	search pop time indexes
 ***********************************/
unsigned long get_time_idx(struct timeval tval_start, struct timeval tval_end, unsigned char unique, long idx[2])
{
	long ini;
	long fin;
	unsigned long result;
	unsigned long count;
	int k=0,i;
	uint32_t last_bunchnr=-1;
	long tmp_idx = this->idx;
	
	struct timeval *time_ptr = (struct timeval *)this->time;
	uint32_t *bunchnr_ptr = (uint32_t *)this->bunchnr;
	
	if(this->counter <= this->size)
	{
		ini = (tmp_idx  - this->counter+this->size)%this->size;	//TODO: ini = 0; ?
		fin =(tmp_idx-1+this->size)%this->size;
	}
	else
	{
		ini = tmp_idx;
		fin =(tmp_idx-1+this->size)%this->size;
	}
	idx[0] = -1;
	idx[1] = -1;
	PRINT("this->idx: %li, this->counter: %li, this->size: %li\n", tmp_idx, this->counter, this->size);
	PRINT("ini: %li\n", ini);
	PRINT("fin: %li\n", fin);

	if((tval_start.tv_sec > tval_end.tv_sec)||((tval_start.tv_sec == tval_end.tv_sec)&&(tval_start.tv_usec > tval_end.tv_usec))){//reverse order
		//idx[0] = this->idx;
		//idx[1] = this->idx;
		idx[0] = -1;
		idx[1] = -1;

		PRINT("reverse order: lower = %d, upper = %d\n", (int)tval_start.tv_sec, (int)tval_end.tv_sec);
		return 0;

	}
	else if((tval_start.tv_sec == tval_end.tv_sec)&&(tval_start.tv_usec == tval_end.tv_usec)){// one sample, one search
		long ind = -1;
		ind = search_time(tval_start, ini, fin);

		PRINT("one sample: tval_start, ini, fin: %d, %li, %li\n", (int)tval_start.tv_sec, ini, fin);

		if(ind!=-1){
			idx[0] = ind;
			idx[1] = ind;

			PRINT("found idx[0] = %li, idx[1] = %li, ind = %li\n", idx[0], idx[1], ind);
			if(idx[1] == ini)	//if returned first index
			{
				if(tv_compare(time_ptr[ini],tval_end) > 0)	//and if tval_end is older than the first present in buffer
				{
					idx[0] = -1;		// -> not found
					idx[1] = -1;		// -> not found
					return 0;
				}
			}

		}
		else
		{
			PRINT("not found \n");
			return 0;
		}

	}
	else{// more samples
		long ind = -1;

		PRINT("right order: tval_start.sec, tval_end.sec, ini, fin: %d, %d, %li, %li\n", (int)tval_start.tv_sec, (int)tval_end.tv_sec, ini, fin);
		PRINT("right order: lower = %d, upper = %d\n", (int)tval_start.tv_sec, (int)tval_end.tv_sec);

		ind = search_time(tval_start, ini, fin);
		if(ind!=-1)
		{
			idx[0] = ind;

			PRINT("found idx[0] = %li\n", idx[0]);
			ini = ind;
		}
		else
		{
			PRINT("not found tval_start\n");
		}

		ind = -1;
		ind = search_time(tval_end, ini, fin);
		if(ind!=-1)
		{
			idx[1] = ind;

			PRINT("found idx[1] = %li\n", idx[1]);

		}
		else
		{
			PRINT("not found tval_end\n");
		}
		
		if(idx[1] == ini && idx[1] == idx[0])	//if tval_end and tval_start returned first index
		{
			if(tv_compare(time_ptr[ini],tval_end) > 0)	//and if tval_end is older than the first present in buffer
			{
				idx[0] = -1;		// -> not found
				idx[1] = -1;		// -> not found
				return 0;
			}
		}

	}
	if(idx[1] > idx[0])
		count = idx[1] - idx[0];
	else if(idx[1] < idx[0])
		count = this->size - idx[0] + idx[1];
	else
		count = 0;
	result = count+1;

	if(unique==1)
	{
		count = 0;
		i=idx[0];
		while(k<result)
		{
			if(last_bunchnr != bunchnr_ptr[i])
			{
				last_bunchnr = bunchnr_ptr[i];
				count++;
			}
			i+=1;
			if(i==this->size)i=0;
			k++;
		}
		result = count;
	}
	return result;
};

/***********************************
 *	search pop bunch indexes
 ***********************************/
unsigned long get_bunch_idx(uint32_t bunch_start, uint32_t bunch_end, unsigned char unique, long idx[2])
{
	long ini,fin;
	unsigned long result,count;
	int k=0,i;
	uint32_t last_bunchnr=-1;
	long tmp_idx = this->idx;
	uint32_t *bunchnr_ptr = (uint32_t *)&bunchnr[0];
	
	if(this->counter <= this->size)
	{
		ini = (tmp_idx  - this->counter+this->size)%this->size; //TODO: ini = 0; ?
		fin =(tmp_idx-1+this->size)%this->size;
	}
	else
	{
		ini = tmp_idx;
		fin =(tmp_idx-1+this->size)%this->size;
	}
	idx[0] = -1;
	idx[1] = -1;
	//PRINT("buffer=0x%lx\n", (unsigned long)buffer);	//TEMP!!
	PRINT("this->idx: %li, this->counter: %li, this->size: %li\n", tmp_idx, this->counter, this->size);
	PRINT("ini: %li\n", ini);
	PRINT("fin: %li\n", fin);

	if(bunch_start > bunch_end)
	{//reverse order
		//idx[0] = this->idx;
		//idx[1] = this->idx;
		idx[0] = -1;
		idx[1] = -1;
		
		PRINT("reverse order: lower = %li, upper = %li\n", bunch_start, bunch_end);
		return 0;
	}
	else if(bunch_start == bunch_end)
	{// one sample, one search
		long ind = -1;
		ind = search_limit(bunch_start, ini, fin, unique);

		PRINT("one sample: bunch_start, ini, fin: %li, %li, %li\n", bunch_start, ini, fin);
		idx[0] = ind;
		idx[1] = ind;
		if(ind!=-1)
		{
			PRINT("found idx[0] = %li, idx[1] = %li, ind = %li\n", idx[0], idx[1], ind);
			if((bunchnr_ptr[ini] > bunch_end) && (unique != 2))	//and if bunch_end is older than the first present in buffer
			{
				idx[0] = -1;		// -> not found
				idx[1] = -1;		// -> not found
				return 0;
			}
		}
		else
		{
			PRINT("not found \n");
		}
		if((idx[0] == -1) || (idx[1] == -1))
			return 0;

	}
	else
	{// more samples
		long ind = -1;

		PRINT("right order: bunch_start, bunch_end, ini, fin: %li, %li, %li, %li\n", bunch_start, bunch_end, ini, fin);
		PRINT("right order: lower = %li, upper = %li\n", bunch_start, bunch_end);

		ind = search_limit(bunch_start, ini, fin, unique);
		idx[0] = ind;
		if(ind!=-1){

			PRINT("found idx[0] = %li\n", idx[0]);
			ini = ind;
		}
		else
		{
			PRINT("not found bunch_start\n");
		}
			
		ind = search_limit(bunch_end, ini, fin, unique);
		idx[1] = ind;
		if(ind!=-1){
			PRINT("found idx[1] = %li\n", idx[1]);
		}
		else
		{
			PRINT("not found bunch_end\n");
		}
			
		if(idx[1] == ini && idx[1] == idx[0])	//if bunch_end and bunch_start returned first index
		{
			if((bunchnr_ptr[ini] > bunch_end) && (unique != 2))	//and if bunch_end is older than the first present in buffer
			{
				idx[0] = -1;		// -> not found
				idx[1] = -1;		// -> not found
				return 0;
			}
		}
		if((idx[0] == -1) || (idx[1] == -1))
			return 0;

	}
	if(idx[1] > idx[0])
		count = idx[1] - idx[0];
	else if(idx[1] < idx[0])
		count = this->size - idx[0] + idx[1];
	else
		count = 0;
	result = count+1;

	if(unique==1)
	{
		count = 0;
		i=idx[0];
		while(k<result)
		{
			if(last_bunchnr != bunchnr_ptr[i])
			{
				last_bunchnr = bunchnr_ptr[i];
				count++;
			}
			i+=1;
			if(i==this->size)i=0;
			k++;
		}
		result = count;
	}
	return result;
};

/***********************************
 *	search pop last indexes 
 ***********************************/
unsigned long get_last_idx(unsigned long last_samples, unsigned char unique, long idx[2])
{
	unsigned long result;
	int k=0,i,j=0,z;
	uint32_t last_bunchnr=-1;
	long tmp_idx = this->idx;
	uint32_t *bunchnr_ptr = (uint32_t *)&bunchnr[0];
	PRINT("requested %li values, pushed %li values, size=%li\n",last_samples, this->counter, this->size);
	PRINT("type=%d idx=%li mode=%li fill_hole=%li\n",this->type, tmp_idx,this->mode,this->fill_hole);
	//PRINT("buff=0x%lx data=0x%lx time=0x%lx bunchnr=0x%lx\n",(unsigned long)buffer, (unsigned long)this->data,(unsigned long)this->time,(unsigned long)this->bunchnr);

	if(last_samples > this->size)
		last_samples = this->size;
	if(last_samples > this->counter)
		last_samples = this->counter;
	if(this->counter != 0)
	{
		if(unique!=1)
		{
			idx[1] = (tmp_idx-1+this->size)%this->size;
			idx[0] = (tmp_idx - last_samples + this->size)%this->size;
		}
		else
		{
			i = tmp_idx;
			//look for the first value with this bunchnumber
			do
			{
				i = (i-1+this->size)%this->size;
				j = (i-1+this->size)%this->size;
				k++;
				if(k>this->counter)
				{
					idx[1]=i;
					idx[0]=i;
					return 1;
				}
			}
			while(bunchnr_ptr[i] == bunchnr_ptr[j]);

			idx[1] = i;
			last_bunchnr = bunchnr_ptr[i];
			j=1;
			k=1;
			//go back of last_samples bunch numbers
			while(k<last_samples)
			{
				j++;
				if(j > this->counter)
					break;
				i = (i-1+this->size)%this->size;
				if(last_bunchnr != bunchnr_ptr[i])
				{
					last_bunchnr = bunchnr_ptr[i];
					k++;
				}
			}
			z = j;
			//look for the first value with this bunchnumber
			j = (i-1+this->size)%this->size;
			while(bunchnr_ptr[i] == bunchnr_ptr[j])
			{
				i = (i-1+this->size)%this->size;
				j = (i-1+this->size)%this->size;
				z++;
				if(z>this->counter)
				{
					idx[1]=i;
					idx[0]=i;
					return 1;
				}
			}
			idx[0] = i;
			return k;
		}
	}
	else
	{
		idx[0] = -1;
		idx[1] = -1;
		return 0;
	}
	if(idx[1] > idx[0])
		result = idx[1] - idx[0];
	else if(idx[1] < idx[0])
		result = this->size - idx[0] + idx[1];
	else
		result = 0;
	return result+1;
};

/***********************************
 *	set method for mode
 ***********************************/
int set_mode(unsigned long mod)
{
	switch(mod){
		case BUF_RT_LINEAR: //buffer not circular
		case BUF_RT_CIRCULAR: //buffer circular
			this->mode = mod;
			return 0;
		break;
		default: //unexpected value
			PRINT("value not allowed\n");
			return -1;
		break;
	}
};

/***********************************
 *	get method for mode
 ***********************************/
ulong get_mode()
{
	return this->mode;
};

/***********************************
 *	get method for size
 ***********************************/
long get_size()
{
	return this->size;
};

/***********************************
 *	get data size
 ***********************************/
long get_data_size()
{
	return get_data_size(identity<T>());
};
private:
template<typename TL>
long get_data_size(identity<TL>)
{
	return (size > 0) ? 1 : 0;
}
template<typename TL>
long get_data_size(identity<_data_array<TL> >)
{
	return (size > 0) ? data[0].get_count() : 0;
}

public:

/***********************************
 *	set method for fill_hole
 ***********************************/
int set_fill_hole(unsigned long fill){

	switch(fill){
		case BUF_RT_NO_FILL: // 0 = no fill
		case BUF_RT_FILL_1: // 1 = fill with 0xffffffff or NaN
		case BUF_RT_FILL_PREV: // 2 = fill with previous value inside the buffer
		case BUF_RT_FILL_0: // 3 = fill with 0
			this->fill_hole = fill;
			return 0;
		break;
		default: //unexpected value

			PRINT("value not allowed\n");

			return -1;
		break;
	}
};

/***********************************
 *	get method for fill_hole
 ***********************************/
ulong get_fill_hole(){
	return this->fill_hole;
};

/***********************************
 *	get method for type
 ***********************************/
/*type_t get_type()
{
	switch(this->type)
	{
		case typeChar:
			PRINT("type is char: %d\n",this->type);
			break;
		case typeShort:
			PRINT("type is short: %d\n",this->type);
			break;
		case typeLong:
			PRINT("type is long: %d\n",this->type);
			break;
		case typeFloat:
			PRINT("type is float: %d\n",this->type);
			break;
		case typeDouble:
			PRINT("type is double: %d\n",this->type);
			break;
		case typeLongLong:
			PRINT("type is long long: %d\n",this->type);
			break;
		default:
			break;
		}
	return this->type;
};*/

/***********************************
 *	search method 
 ***********************************/
long search_limit(uint32_t nr, long lower, long upper, unsigned char unique)
{
	int found = -1;
	long ini = lower;
	long fin = upper;
	long med = (ini+((fin-ini+this->size)%this->size)/2)%this->size;
	long result = med;
	int counter = 0;

	uint32_t *bunchnr_ptr = (uint32_t *)this->bunchnr;

	PRINT("searching from ini = %li, to fin = %li\n", ini, fin);
	if((ini==fin) && (bunchnr_ptr[ini] != nr))
	{
		if(unique != 2)
			return -1;
		else
			return ini;
	}
	while(found<0)
	{
		if(bunchnr_ptr[med] == nr)
		{	//equals
			found = 0;
			while(bunchnr_ptr[(med-1+this->size)%this->size]==nr && med >= lower)
			{
				med = (med-1+this->size)%this->size;
				PRINT("changing med to %li\n", med);
			}
			PRINT("found med=%li, bunchnr[med]=%li, med-1=%li, bunchnr[med-1]=%li\n", med, bunchnr_ptr[med], (med-1+this->size)%this->size, bunchnr_ptr[(med-1+this->size)%this->size]);
			result = med;
		}
		else
		{
			if(((fin-ini+this->size)%this->size)==1)
			{
				//PRINT("in step=1: ini=%li, fin=%li bunchnr[ini]=%li bunchnr[fin]=%li\n", ini, fin ,bunchnr_ptr[ini], bunchnr_ptr[fin]);
				//step == 1, last test
				if(bunchnr_ptr[ini]==nr)
				{
					found = 0;
					while(bunchnr_ptr[(ini-1+this->size)%this->size]==nr && ini >= lower)
					{
						ini = (ini-1+this->size)%this->size;
						PRINT("changing ini to %li\n", ini);
					}
					PRINT("found ini=%li, bunchnr[ini]=%li, ini-1=%li, bunchnr[ini-1]=%li\n", ini, bunchnr_ptr[ini], (ini-1+this->size)%this->size, bunchnr_ptr[(ini-1+this->size)%this->size]);
					result = ini;
				}
				else if(bunchnr_ptr[ini]>nr)
				{
					found = 0;
					if(unique != 2)
						result = -1;
					else
						result = ini;
				}
				else if(bunchnr_ptr[fin]==nr)
				{ 
					found = 0;
					while(bunchnr_ptr[(fin-1+this->size)%this->size]==nr && fin >= lower)
					{
						fin = (fin-1+this->size)%this->size;
						PRINT("changing fin to %li\n", fin);
					}
					PRINT("found fin=%li, bunchnr[fin]=%li, fin-1=%li, bunchnr[fin-1]=%li\n", fin, bunchnr_ptr[fin], (fin-1+this->size)%this->size, bunchnr_ptr[(fin-1+this->size)%this->size]);
					result = fin;
				}
				else if(bunchnr_ptr[fin]<nr)
				{ 
					found = 0;
					if(unique != 2)
						result = -1;
					else
						result = fin;
				}
				//between ini and fin, but step == 1, hole?
				else
				{
					found = 0;
					//result = ini;
					if(unique != 2)
						result = -1;
					else
						result = ini;
				}
				break;
			}
			med = (ini+((fin-ini+this->size)%this->size)/2)%this->size;
			if(bunchnr_ptr[med] < nr)
			{	//lower
				ini = med/*%this->size*/;
				//PRINT("adjusted ini: ini=%li, fin=%li bunchnr[ini]=%li\n", ini, fin, bunchnr_ptr[ini]);
			}
			else
			{	//greater
				fin = med/*%this->size*/;
				//PRINT("adjusted fin: ini=%li, fin=%li bunchnr[fin]=%li\n", ini, fin, bunchnr_ptr[fin]);
			}
		}
		counter+=1;	
	}

	PRINT("Result: idx=%li bunchnr[idx]=%li bunchnr[lower]=%li bunchnr[upper]=%li\n", result, (result != -1) ? bunchnr_ptr[result] : -1, bunchnr_ptr[lower], bunchnr_ptr[upper]);

	return result;
};

int tv_compare(struct timeval ts1, struct timeval ts2)
{
	if(ts1.tv_sec < ts2.tv_sec)
		return - 1;
	else if(ts1.tv_sec == ts2.tv_sec)
	{
		if(ts1.tv_usec < ts2.tv_usec)
			return -1;
		if(ts1.tv_usec == ts2.tv_usec)
			return 0;
		else
			return 1;
	}
	else
		return 1;

}

long search_time(struct timeval p, long lower, long upper)
{
	int found = -2;
	long ini = lower;
	long fin = upper;
	long med = (ini+((fin-ini+this->size)%this->size)/2)%this->size;
	long result = med;
//	long reslow = med;
//	long reshi = med;
	int counter = 0;
//	int counter2 = 0;

	struct timeval *time_ptr = (struct timeval *)this->time;


	PRINT("searching from ini = %li, to fin = %li\n", ini, fin);
#if 0
	while(found<0){
		if(time_ptr[med].tv_sec == p.tv_sec){	//equals
			found +=1;
			reslow = med;
			reshi = med;
			while(time_ptr[reslow].tv_sec == p.tv_sec){
				reslow=(reslow-1+this->size)%this->size;
			}
			while(time_ptr[reshi].tv_sec == p.tv_sec){
				reshi = (reshi+1)%this->size;
			}
			reslow = (reslow+1)%this->size;
			reshi = (reshi-1+this->size)%this->size;
		}
		else{
			if(((fin-ini+this->size)%this->size)==1)
			{
				if(time_ptr[fin].tv_sec!=p.tv_sec){
					found +=1;
					reslow = ini;
				}
				else{ 
					found +=1;
					reshi = fin;
				}
			}
			med = (ini+((fin-ini+this->size)%this->size)/2)%this->size;
			if(time_ptr[med].tv_sec < p.tv_sec){	//lower
				ini = med/*%this->size*/;
			}
			else{	//greater
				fin = med/*%this->size*/;
			}
		}
		counter+=1;	
	}

	PRINT("reslow %li reshi %li\n", reslow, reshi);

	
	//finding usec
	found = -1;

	PRINT("searching from reslow = %li, to reshi = %li\n", reslow, reshi);

	med = reslow+((reshi-reslow+this->size)%this->size)/2;
	result = med;
	while(found<0){
		if(time_ptr[med].tv_usec == p.tv_usec){	//equals
			found = 0;
			while(time_ptr[med].tv_usec == p.tv_usec){
				med=(med-1+this->size)%this->size;
			}
			result =(med+1)%this->size;
		}
		else{
			if(((reshi-reslow+this->size)%this->size)==1)
			{
				if(time_ptr[reshi].tv_usec!=p.tv_usec){
					found = 0;
					result =  reslow;
				}
				else{ 
					found = 0;
					result = reshi;
				}
			}
			med = reslow+((reshi-reslow+this->size)%this->size)/2;
			if(time_ptr[med].tv_usec < p.tv_usec){	//lower
				reslow = med%this->size;
			}
			else{	//greater
				reshi = med%this->size;
			}
		}
		counter2+=1;	
	}
#else
	if((ini==fin) && (tv_compare(time_ptr[ini],p)!=0))
		return -1;
	while(found<0){
		if(tv_compare(time_ptr[med],p)==0){	//equals
			found = 0;
			result = med;
		}
		else{
			if(((fin-ini+this->size)%this->size)==1)
			{
				//step == 1, last test
				if(tv_compare(time_ptr[ini],p)>=0){	//equal to p, or p lower than the min
					found = 0;
					result = ini;
				}
				else if(tv_compare(time_ptr[fin],p)<=0){	//equal to p, or p greater than the max
					found = 0;
					result = fin;
				}
				//between ini and fin, but step == 1
				else{
					found = 0;
					result = ini;
				}
				break;
			}
			med = (ini+((fin-ini+this->size)%this->size)/2)%this->size;
			if(tv_compare(time_ptr[med],p)<0){	//lower
				ini = med/*%this->size*/;
			}
			else{	//greater
				fin = med/*%this->size*/;
			}
		}
		counter+=1;	
	}
#endif
	PRINT("Searched %d.%d, result %li time[lower]=%d.%d time[upper]=%d.%d\n", (int)p.tv_sec, (int)p.tv_usec, result, (int)time_ptr[lower].tv_sec, (int)time_ptr[lower].tv_usec, (int)time_ptr[upper].tv_sec, (int)time_ptr[upper].tv_usec);

	
	return result;
	
};

};
}	//namespace tbuffer

#endif //__TBUFFER_H
